from __future__ import annotations

import os
import shutil
import sys
from pathlib import Path

# setuptools *must* come before Cython, otherwise Cython's distutils hacking
# will override setuptools' build_ext command and cause problems in other build
# systems such as nix.

from setuptools import Distribution, Extension
from Cython.Build import build_ext, cythonize

SOURCE_DIR = Path("koerce")
BUILD_DIR = Path("cython_build")


extensions = [
    Extension("koerce.annots", ["koerce/annots.py"]),
    Extension("koerce.builders", ["koerce/builders.py"]),
    Extension("koerce.patterns", ["koerce/patterns.py"]),
    # Extension("koerce.utils", ["koerce/utils.py"]),
]

cythonized_modules = cythonize(
    extensions,
    build_dir=BUILD_DIR,
    # generate anannotated .html output files.
    annotate=True,
    compiler_directives={
        "language_level": "3",
        "binding": False,
        "boundscheck": False,
        "nonecheck": False,
        "always_allow_keywords": False,
    },
)

dist = Distribution({"ext_modules": cythonized_modules})
cmd = build_ext(dist)
cmd.ensure_finalized()
cmd.run()

# remove all binaries *.so, *.dll, *.lib, *.pyd, *.dylib
for ext in ["so", "dll", "lib", "pyd", "dylib"]:
    for f in SOURCE_DIR.glob(f"*.{ext}"):
        f.unlink()
        print(f"removed {f!r}", file=sys.stderr)  # noqa: T201

for output in cmd.get_outputs():
    relative_extension = os.path.relpath(output, cmd.build_lib)
    print(f"{output} -> {relative_extension}", file=sys.stderr)  # noqa: T201
    shutil.copyfile(output, relative_extension)



from __future__ import annotations

import importlib
from collections.abc import Callable, Mapping, Sequence
from enum import Enum
from inspect import Parameter, Signature
from types import UnionType
from typing import (
    Annotated,
    Any,
    ForwardRef,
    Literal,
    Optional,
    TypeVar,
    Union,
)

import cython
from typing_extensions import GenericMeta, Self, get_original_bases

# TODO(kszucs): would be nice to cimport Signature and Builder
from .builders import Builder, Deferred, Variable, builder
from .utils import (
    RewindableIterator,
    get_type_args,
    get_type_boundvars,
    get_type_origin,
    get_type_params,
)


class CoercionError(Exception):
    pass


Context = dict[str, Any]


@cython.final
@cython.cclass
class NoMatchError(Exception):
    pass


@cython.cclass
class NoMatch:
    def __init__(self):
        raise ValueError("Cannot instantiate NoMatch")


@cython.cclass
class Pattern:
    @staticmethod
    def from_typehint(
        annot: Any, allow_coercion: bool = True, self_qualname: Optional[str] = None
    ) -> Pattern:
        """Construct a validator from a python type annotation.

        Parameters
        ----------
        annot
            The typehint annotation to construct the pattern from. This must be
            an already evaluated type annotation.
        allow_coercion
            Whether to use coercion if the typehint is a Coercible type.
        self_qualname
            The qualname of the class to use for Self typehints.

        Returns
        -------
        A pattern that matches the given type annotation.
        """
        args: tuple = get_type_args(annot)
        origin: Any = get_type_origin(annot)
        options: dict[str, Any] = {
            "allow_coercion": allow_coercion,
            "self_qualname": self_qualname,
        }

        if origin is None:
            # the typehint is not generic
            if annot is Ellipsis or annot is Any:
                # treat both `Any` and `...` as wildcard
                return _any
            elif annot is Self:
                if self_qualname is None:
                    raise ValueError("self_qualname must be provided for Self typehint")
                return LazyInstanceOf(self_qualname)
            elif isinstance(annot, type):
                # the typehint is a concrete type (e.g. int, str, etc.)
                if allow_coercion and hasattr(annot, "__coerce__"):
                    # the type implements the Coercible protocol so we try to
                    # coerce the value to the given type rather than checking
                    return CoercedTo(annot)
                else:
                    return InstanceOf(annot)
            elif isinstance(annot, TypeVar):
                # if the typehint is a type variable we try to construct a
                # validator from it only if it is covariant and has a bound
                if not annot.__covariant__:
                    raise NotImplementedError(
                        "Only covariant typevars are supported for now"
                    )
                if annot.__bound__:
                    return Pattern.from_typehint(annot.__bound__, **options)
                else:
                    return _any
            elif isinstance(annot, Enum):
                # for enums we check the value against the enum values
                return EqValue(annot)
            elif isinstance(annot, str):
                # for strings and forward references we check in a lazy way
                return LazyInstanceOf(annot)
            elif isinstance(annot, ForwardRef):
                return LazyInstanceOf(annot.__forward_arg__)
            else:
                raise TypeError(f"Cannot create validator from annotation {annot!r}")
        # elif origin is CoercedTo:
        #     return CoercedTo(args[0])
        elif origin is Literal:
            # for literal types we check the value against the literal values
            return IsIn(args)
        elif origin is UnionType or origin is Union:
            # this is slightly more complicated because we need to handle
            # Optional[T] which is Union[T, None] and Union[T1, T2, ...]
            *rest, last = args
            if last is type(None):
                # the typehint is Optional[*rest] which is equivalent to
                # Union[*rest, None], so we construct an Option pattern
                if len(rest) == 1:
                    inner = Pattern.from_typehint(rest[0], **options)
                else:
                    variants = [Pattern.from_typehint(arg, **options) for arg in rest]
                    inner = AnyOf(*variants)
                return Option(inner)
            else:
                # the typehint is Union[*args] so we construct an AnyOf pattern
                variants = [Pattern.from_typehint(arg, **options) for arg in args]
                return AnyOf(*variants)
        elif origin is Annotated:
            # the Annotated typehint can be used to add extra validation logic
            # to the typehint, e.g. Annotated[int, Positive], the first argument
            # is used for isinstance checks, the rest are applied in conjunction
            annot, *extras = args
            return AllOf(
                Pattern.from_typehint(annot, **options),
                *extras,
            )
        elif origin is Callable:
            # the Callable typehint is used to annotate functions, e.g. the
            # following typehint annotates a function that takes two integers
            # and returns a string: Callable[[int, int], str]
            if args:
                # callable with args and return typehints construct a special
                # CallableWith validator
                arg_hints, return_hint = args
                arg_patterns = [
                    Pattern.from_typehint(arg, **options) for arg in arg_hints
                ]
                return_pattern = Pattern.from_typehint(return_hint, **options)
                return CallableWith(arg_patterns, return_pattern)
            else:
                # in case of Callable without args we check for the Callable
                # protocol only
                return InstanceOf(Callable)
        elif issubclass(origin, tuple):
            # construct validators for the tuple elements, but need to treat
            # variadic tuples differently, e.g. tuple[int, ...] is a variadic
            # tuple of integers, while tuple[int] is a tuple with a single int
            first, *rest = args
            if rest == [Ellipsis]:
                return TupleOf(Pattern.from_typehint(first, **options))
            else:
                patterns = [Pattern.from_typehint(arg, **options) for arg in args]
                return PatternList(patterns, origin)
        elif issubclass(origin, Sequence):
            # construct a validator for the sequence elements where all elements
            # must be of the same type, e.g. Sequence[int] is a sequence of ints
            value_inner = Pattern.from_typehint(args[0], **options)
            return SequenceOf(value_inner, type_=origin)
        elif issubclass(origin, Mapping):
            # construct a validator for the mapping keys and values, e.g.
            # Mapping[str, int] is a mapping with string keys and int values
            key_inner = Pattern.from_typehint(args[0], **options)
            value_inner = Pattern.from_typehint(args[1], **options)
            return MappingOf(key_inner, value_inner, origin)
        elif isinstance(origin, GenericMeta):
            # construct a validator for the generic type, see the specific
            # Generic* validators for more details
            if allow_coercion and hasattr(origin, "__coerce__") and args:
                return GenericCoercedTo(annot)
            else:
                return GenericInstanceOf(annot)
        else:
            raise TypeError(
                f"Cannot create validator from annotation {annot!r} {origin!r}"
            )

    def apply(self, value, context=None):
        if context is None:
            context = {}
        try:
            return self.match(value, context)
        except NoMatchError:
            return NoMatch

    @cython.cfunc
    def match(self, value, ctx: Context): ...

    def __repr__(self) -> str: ...

    def __eq__(self, other) -> bool:
        return type(self) is type(other) and self.equals(other)

    def __invert__(self) -> Not:
        """Syntax sugar for matching the inverse of the pattern."""
        return Not(self)

    def __or__(self, other: Pattern) -> AnyOf:
        """Syntax sugar for matching either of the patterns.

        Parameters
        ----------
        other
            The other pattern to match against.

        Returns
        -------
        New pattern that matches if either of the patterns match.
        """
        if isinstance(other, AnyOf):
            return AnyOf(self, *cython.cast(AnyOf, other).inners)
        else:
            return AnyOf(self, other)

    def __and__(self, other: Pattern) -> AllOf:
        """Syntax sugar for matching both of the patterns.

        Parameters
        ----------
        other
            The other pattern to match against.

        Returns
        -------
        New pattern that matches if both of the patterns match.
        """
        if isinstance(other, AllOf):
            return AllOf(self, *cython.cast(AllOf, other).inners)
        else:
            return AllOf(self, other)

    def __rmatmul__(self, name) -> Capture:
        """Syntax sugar for capturing a value.

        Parameters
        ----------
        name
            The name of the capture.

        Returns
        -------
        New capture pattern.

        """
        return Capture(name, self)

    def __rshift__(self, other: Deferred) -> Replace:
        """Syntax sugar for replacing a value.

        Parameters
        ----------
        other
            The deferred to use for constructing the replacement value.

        Returns
        -------
        New replace pattern.

        """
        return Replace(self, other)

    def __iter__(self) -> SomeOf:
        yield SomeOf(self)


@cython.final
@cython.cclass
class Anything(Pattern):
    def __repr__(self) -> str:
        return "Anything()"

    def equals(self, other: Anything) -> bool:
        return True

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        return value


_any = Anything()


@cython.final
@cython.cclass
class Nothing(Pattern):
    def __repr__(self) -> str:
        return "Nothing()"

    def equals(self, other: Nothing) -> bool:
        return True

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        raise NoMatchError()


@cython.final
@cython.cclass
class IdenticalTo(Pattern):
    value: Any

    def __init__(self, value):
        self.value = value

    def __repr__(self) -> str:
        return f"IdenticalTo({self.value!r})"

    def equals(self, other: IdenticalTo) -> bool:
        return self.value == other.value

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        if value is self.value:
            return value
        else:
            raise NoMatchError()


@cython.ccall
def Eq(value) -> Pattern:
    if isinstance(value, (Deferred, Builder)):
        return EqDeferred(value)
    else:
        return EqValue(value)


@cython.final
@cython.cclass
class EqValue(Pattern):
    value: Any

    def __init__(self, value: Any):
        self.value = value

    def __repr__(self) -> str:
        return f"EqValue({self.value!r})"

    def equals(self, other: EqValue) -> bool:
        return self.value == other.value

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        if value == self.value:
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class EqDeferred(Pattern):
    """Pattern that checks a value equals to the given value.

    Parameters
    ----------
    value
        The value to check against.

    """

    value: Builder

    def __init__(self, value):
        self.value = builder(value)

    def __repr__(self) -> str:
        return f"EqDeferred({self.value!r})"

    @cython.cfunc
    def match(self, value, ctx: Context):
        # ctx["_"] = value
        # TODO(kszucs): Builder is not cimported so self.value.build() cannot be
        # used, hence using .apply() instead
        if value == self.value.apply(ctx):
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class TypeOf(Pattern):
    type_: Any

    def __init__(self, type_):
        assert isinstance(type_, type)
        self.type_ = type_

    def __repr__(self) -> str:
        return f"TypeOf({self.type_!r})"

    def equals(self, other: TypeOf) -> bool:
        return self.type_ == other.type_

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        if type(value) is self.type_:
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class InstanceOf(Pattern):
    # performance doesn't seem to be affected:
    # https://github.com/kszucs/koerce/pull/6#discussion_r1705833034
    type_: Any

    def __init__(self, type_: Any):
        self.type_ = type_

    def __repr__(self) -> str:
        return f"InstanceOf({self.type_!r})"

    def __call__(self, *args, **kwargs):
        return ObjectOf(self.type_, *args, **kwargs)

    def equals(self, other: InstanceOf) -> bool:
        return self.type_ == other.type_

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        if isinstance(value, self.type_):
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class LazyInstanceOf(Pattern):
    qualname: str
    package: str
    type_: Any

    def __init__(self, qualname: str):
        if not isinstance(qualname, str):
            raise TypeError("qualname must be a string")
        _common_package_aliases: dict[str, str] = {
            "pa": "pyarrow",
            "pd": "pandas",
            "np": "numpy",
            "tf": "tensorflow",
        }
        package: str = qualname.split(".")[0]
        self.qualname = qualname
        self.package = _common_package_aliases.get(package, package)
        self.type_ = None

    def __repr__(self) -> str:
        return f"LazyInstanceOf({self.qualname!r})"

    def __call__(self, *args, **kwargs):
        return ObjectOf(self, *args, **kwargs)

    def equals(self, other: LazyInstanceOf) -> bool:
        return self.qualname == other.qualname

    @cython.cfunc
    def _import_type(self):
        module_name, type_name = self.qualname.rsplit(".", 1)
        module = importlib.import_module(module_name)
        try:
            self.type_ = getattr(module, type_name)
        except AttributeError:
            raise ImportError(f"Could not import {type_name} from {module_name}")

    @cython.cfunc
    def match(self, value, ctx: Context):
        if self.type_ is not None:
            if isinstance(value, self.type_):
                return value
            else:
                raise NoMatchError()

        klass: Any
        package: str
        for klass in type(value).__mro__:
            package = klass.__module__.split(".", 1)[0]
            if package == self.package:
                self._import_type()
                if isinstance(value, self.type_):
                    return value
                else:
                    raise NoMatchError()

        raise NoMatchError()


@cython.ccall
def GenericInstanceOf(typ) -> Pattern:
    # TODO(kszucs): could be expressed using ObjectOfN..
    nparams: int = len(get_type_params(typ))
    if nparams == 1:
        return GenericInstanceOf1(typ)
    elif nparams == 2:
        return GenericInstanceOf2(typ)
    else:
        return GenericInstanceOfN(typ)


@cython.final
@cython.cclass
class GenericInstanceOf1(Pattern):
    origin: Any
    name1: str
    pattern1: Pattern

    def __init__(self, typ):
        self.origin = get_type_origin(typ)

        ((self.name1, type1),) = get_type_boundvars(typ).items()
        self.pattern1 = Pattern.from_typehint(type1, allow_coercion=False)

    def __repr__(self) -> str:
        return f"GenericInstanceOf1({self.origin!r}, name1={self.name1!r}, pattern1={self.pattern1!r})"

    def __call__(self, *args, **kwargs):
        return ObjectOf(self, *args, **kwargs)

    def equals(self, other: GenericInstanceOf1) -> bool:
        return (
            self.origin == other.origin
            and self.name1 == other.name1
            and self.pattern1 == other.pattern1
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.origin):
            raise NoMatchError()

        attr1 = getattr(value, self.name1)
        self.pattern1.match(attr1, ctx)

        return value


@cython.final
@cython.cclass
class GenericInstanceOf2(Pattern):
    origin: Any
    name1: str
    name2: str
    pattern1: Pattern
    pattern2: Pattern

    def __init__(self, typ):
        self.origin = get_type_origin(typ)

        (self.name1, type1), (self.name2, type2) = get_type_boundvars(typ).items()
        self.pattern1 = Pattern.from_typehint(type1, allow_coercion=False)
        self.pattern2 = Pattern.from_typehint(type2, allow_coercion=False)

    def __repr__(self) -> str:
        return f"GenericInstanceOf2({self.origin!r}, name1={self.name1!r}, pattern1={self.pattern1!r}, name2={self.name2!r}, pattern2={self.pattern2!r})"

    def __call__(self, *args, **kwargs):
        return ObjectOf(self, *args, **kwargs)

    def equals(self, other: GenericInstanceOf2) -> bool:
        return (
            self.origin == other.origin
            and self.name1 == other.name1
            and self.pattern1 == other.pattern1
            and self.name2 == other.name2
            and self.pattern2 == other.pattern2
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.origin):
            raise NoMatchError()

        attr1 = getattr(value, self.name1)
        self.pattern1.match(attr1, ctx)

        attr2 = getattr(value, self.name2)
        self.pattern2.match(attr2, ctx)

        return value


@cython.final
@cython.cclass
class GenericInstanceOfN(Pattern):
    origin: Any
    fields: dict[str, Pattern]

    def __init__(self, typ):
        self.origin = get_type_origin(typ)

        name: str
        self.fields = {}
        for name, type_ in get_type_boundvars(typ).items():
            self.fields[name] = Pattern.from_typehint(type_, allow_coercion=False)

    def __repr__(self) -> str:
        return f"GenericInstanceOfN({self.origin!r}, fields={self.fields!r})"

    def __call__(self, *args, **kwargs):
        return ObjectOf(self, *args, **kwargs)

    def equals(self, other: GenericInstanceOfN) -> bool:
        return self.origin == other.origin and self.fields == other.fields

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.origin):
            raise NoMatchError()

        name: str
        pattern: Pattern
        for name, pattern in self.fields.items():
            attr = getattr(value, name)
            pattern.match(attr, ctx)

        return value


@cython.final
@cython.cclass
class SubclassOf(Pattern):
    type_: Any

    def __init__(self, type_: Any):
        self.type_ = type_

    def __repr__(self) -> str:
        return f"SubclassOf({self.type_!r})"

    def equals(self, other: SubclassOf) -> bool:
        return self.type_ == other.type_

    @cython.cfunc
    @cython.inline
    def match(self, value, ctx: Context):
        if issubclass(value, self.type_):
            return value
        else:
            raise NoMatchError()


# @cython.ccall
# def As(type_: Any) -> Pattern:
#     origin = get_type_origin(type_)
#     if origin is None:
#         if hasattr(type_, "__coerce__"):
#             return CoercedTo(type_)
#         else:
#             return AsType(type_)
#     else:
#         if hasattr(origin, "__coerce__"):
#             return GenericCoercedTo(type_)
#         else:
#             return AsType(type_)


@cython.final
@cython.cclass
class AsType(Pattern):
    type_: Any

    def __init__(self, type_: Any):
        self.type_ = type_

    def __repr__(self) -> str:
        return f"AsType({self.type_!r})"

    def equals(self, other: AsType) -> bool:
        return self.type_ == other.type_

    @cython.cfunc
    def match(self, value, ctx: Context):
        try:
            return self.type_(value)
        except ValueError:
            raise NoMatchError()


@cython.final
@cython.cclass
class CoercedTo(Pattern):
    type_: Any

    def __init__(self, type_: Any):
        if not hasattr(type_, "__coerce__"):
            raise TypeError(f"{type_} does not implement the Coercible protocol")
        self.type_ = type_

    def __repr__(self) -> str:
        return f"CoercedTo({self.type_!r})"

    def __call__(self, *args, **kwargs):
        return ObjectOf(self, *args, **kwargs)

    def equals(self, other: CoercedTo) -> bool:
        return self.type_ == other.type_

    @cython.cfunc
    def match(self, value, ctx: Context):
        try:
            value = self.type_.__coerce__(value)
        except CoercionError:
            raise NoMatchError()

        if isinstance(value, self.type_):
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class GenericCoercedTo(Pattern):
    origin: Any
    params: dict[str, type]
    checker: Pattern

    def __init__(self, typ):
        self.origin = get_type_origin(typ)
        if not hasattr(self.origin, "__coerce__"):
            raise TypeError(f"{self.origin} does not implement the Coercible protocol")
        self.checker = GenericInstanceOf(typ)

        # get all type parameters for the generic class in its type hierarchy
        self.params = {}
        for base in get_original_bases(self.origin):
            self.params.update(get_type_params(base))
        self.params.update(get_type_params(typ))

    def __repr__(self) -> str:
        return f"GenericCoercedTo({self.origin!r}, params={self.params!r})"

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return ObjectOf(self, *args, **kwds)

    def equals(self, other: GenericCoercedTo) -> bool:
        return self.origin == other.origin and self.params == other.params

    @cython.cfunc
    def match(self, value, ctx: Context):
        try:
            value = self.origin.__coerce__(value, **self.params)
        except CoercionError:
            raise NoMatchError()

        self.checker.match(value, ctx)

        return value


@cython.final
@cython.cclass
class Not(Pattern):
    inner: Pattern

    def __init__(self, inner):
        self.inner = pattern(inner)

    def __repr__(self) -> str:
        return f"Not({self.inner!r})"

    def equals(self, other: Not) -> bool:
        return self.inner == other.inner

    @cython.cfunc
    def match(self, value, ctx: Context):
        try:
            self.inner.match(value, ctx)
        except NoMatchError:
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class AnyOf(Pattern):
    inners: list[Pattern]

    def __init__(self, *inners: Pattern):
        self.inners = [pattern(inner) for inner in inners]

    def __repr__(self) -> str:
        return f"AnyOf({self.inners!r})"

    def equals(self, other: AnyOf) -> bool:
        return self.inners == other.inners

    @cython.cfunc
    def match(self, value, ctx: Context):
        inner: Pattern
        for inner in self.inners:
            try:
                return inner.match(value, ctx)
            except NoMatchError:
                pass
        raise NoMatchError()

    def __or__(self, other: Pattern) -> AnyOf:
        """Syntax sugar for matching either of the patterns.

        Parameters
        ----------
        other
            The other pattern to match against.

        Returns
        -------
        New pattern that matches if either of the patterns match.
        """
        if isinstance(other, AnyOf):
            return AnyOf(*self.inners, *cython.cast(AnyOf, other).inners)
        else:
            return AnyOf(*self.inners, other)


@cython.final
@cython.cclass
class AllOf(Pattern):
    inners: list[Pattern]

    def __init__(self, *inners: Pattern):
        self.inners = [pattern(inner) for inner in inners]

    def __repr__(self) -> str:
        return f"AllOf({self.inners!r})"

    def equals(self, other: AllOf) -> bool:
        return self.inners == other.inners

    @cython.cfunc
    def match(self, value, ctx: Context):
        inner: Pattern
        for inner in self.inners:
            value = inner.match(value, ctx)
        return value

    def __and__(self, other: Pattern) -> AllOf:
        """Syntax sugar for matching both of the patterns.

        Parameters
        ----------
        other
            The other pattern to match against.

        Returns
        -------
        New pattern that matches if both of the patterns match.
        """
        if isinstance(other, AllOf):
            return AllOf(*self.inners, *cython.cast(AllOf, other).inners)
        else:
            return AllOf(*self.inners, other)


def NoneOf(*args) -> Pattern:
    """Match none of the passed patterns."""
    return Not(AnyOf(*args))


@cython.final
@cython.cclass
class Option(Pattern):
    """Pattern that matches `None` or a value that passes the inner validator.

    Parameters
    ----------
    pattern
        The inner pattern to use.

    """

    pattern: Pattern
    default: Any

    def __init__(self, pat, default=None):
        self.pattern = pattern(pat)
        self.default = default

    def __repr__(self) -> str:
        return f"Option({self.pattern!r}, default={self.default!r})"

    def equals(self, other: Option) -> bool:
        return self.pattern == other.pattern and self.default == other.default

    @cython.cfunc
    def match(self, value, ctx: Context):
        if value is None:
            if self.default is None:
                return None
            else:
                return self.default
        else:
            return self.pattern.match(value, ctx)


@cython.ccall
def If(predicate) -> Pattern:
    if isinstance(predicate, (Deferred, Builder)):
        return IfDeferred(predicate)
    elif callable(predicate):
        return IfFunction(predicate)
    else:
        raise TypeError("Predicate must be a callable or a deferred value")


@cython.final
@cython.cclass
class IfFunction(Pattern):
    """Pattern that checks a value against a predicate.

    Parameters
    ----------
    predicate
        The predicate to use.

    """

    predicate: Callable

    def __init__(self, predicate):
        self.predicate = predicate

    def __repr__(self) -> str:
        return f"IfFunction({self.predicate!r})"

    def equals(self, other: IfFunction) -> bool:
        return self.predicate == other.predicate

    @cython.cfunc
    def match(self, value, ctx: Context):
        if self.predicate(value):
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class IfDeferred(Pattern):
    """Pattern that checks a value against a predicate.

    Parameters
    ----------
    predicate
        The predicate to use.

    """

    builder: Builder

    def __init__(self, obj):
        self.builder = builder(obj)

    def __repr__(self) -> str:
        return f"If({self.builder!r})"

    def equals(self, other: IfDeferred) -> bool:
        return self.builder == other.builder

    @cython.cfunc
    def match(self, value, ctx: Context):
        # TODO(kszucs): Builder is not cimported so self.builder.build()
        # is not available, hence using .apply() instead
        ctx["_"] = value
        if self.builder.apply(ctx):
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class IsIn(Pattern):
    """Pattern that matches if a value is in a given set.

    Parameters
    ----------
    haystack
        The set of values that the passed value should be in.

    """

    haystack: frozenset

    def __init__(self, haystack):
        self.haystack = frozenset(haystack)

    def __repr__(self) -> str:
        return f"IsIn({self.haystack})"

    def equals(self, other: IsIn) -> bool:
        return self.haystack == other.haystack

    @cython.cfunc
    def match(self, value, ctx: Context):
        if value in self.haystack:
            return value
        else:
            raise NoMatchError()


@cython.final
@cython.cclass
class SequenceOf(Pattern):
    """Pattern that matches if all of the items in a sequence match a given pattern.

    Specialization of the more flexible GenericSequenceOf pattern which uses two
    additional patterns to possibly coerce the sequence type and to match on
    the length of the sequence.

    Parameters
    ----------
    item
        The pattern to match against each item in the sequence.
    type
        The type to coerce the sequence to. Defaults to tuple.

    """

    item: Pattern
    type_: Pattern

    def __init__(self, item, type_=list):
        self.item = pattern(item)
        if hasattr(type_, "__coerce__"):
            self.type_ = CoercedTo(type_)
        else:
            try:
                type_([])
            except TypeError:
                self.type_ = AsType(list)
            else:
                self.type_ = AsType(type_)

    def __repr__(self) -> str:
        return f"SequenceOf({self.item!r}, type_={self.type_!r})"

    def equals(self, other: SequenceOf) -> bool:
        return self.item == other.item and self.type_ == other.type_

    @cython.cfunc
    def match(self, values, ctx: Context):
        if isinstance(values, (str, bytes)):
            raise NoMatchError()

        # optimization to avoid unnecessary iteration
        if isinstance(self.item, Anything):
            return values

        try:
            it = iter(values)
        except TypeError:
            raise NoMatchError()

        result: list = []
        for item in it:
            result.append(self.item.match(item, ctx))

        return self.type_.match(result, ctx)


def ListOf(item) -> Pattern:
    return SequenceOf(item, list)


def TupleOf(item) -> Pattern:
    return SequenceOf(item, tuple)


@cython.final
@cython.cclass
class MappingOf(Pattern):
    """Pattern that matches if all of the keys and values match the given patterns.

    Parameters
    ----------
    key
        The pattern to match the keys against.
    value
        The pattern to match the values against.
    type
        The type to coerce the mapping to. Defaults to dict.

    """

    __slots__ = ("key", "value", "type")
    key: Pattern
    value: Pattern
    type_: Pattern

    def __init__(self, key: Pattern, value: Pattern, type_=dict):
        self.key = pattern(key)
        self.value = pattern(value)
        if isinstance(type_, type):
            self.type_ = AsType(type_)
        elif hasattr(type_, "__coerce__"):
            self.type_ = CoercedTo(type_)
        else:
            raise TypeError(f"Cannot coerce to container type {type_}")

    def __repr__(self) -> str:
        return f"MappingOf({self.key!r}, {self.value!r}, {self.type_!r})"

    def equals(self, other: MappingOf) -> bool:
        return (
            self.key == other.key
            and self.value == other.value
            and self.type_ == other.type_
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, Mapping):
            raise NoMatchError()

        result = {}
        for k, v in value.items():
            k = self.key.match(k, ctx)
            v = self.value.match(v, ctx)
            result[k] = v

        return self.type_.match(result, ctx)


def DictOf(key, value) -> Pattern:
    return MappingOf(key, value, dict)


@cython.final
@cython.cclass
class Custom(Pattern):
    """Pattern that matches if a custom function returns True.

    Parameters
    ----------
    func
        The function to use for matching.

    """

    func: Callable

    def __init__(self, func):
        self.func = func

    def __repr__(self) -> str:
        return f"Custom({self.func!r})"

    def equals(self, other: Custom) -> bool:
        return self.func == other.func

    @cython.cfunc
    def match(self, value, ctx: Context):
        return self.func(value, **ctx)


@cython.final
@cython.cclass
class Capture(Pattern):
    """Pattern that captures a value in the context.

    Parameters
    ----------
    pattern
        The pattern to match against.
    key
        The key to use in the context if the pattern matches.

    """

    key: str
    what: Pattern

    def __init__(self, key: str | Deferred | Builder, what=_any):
        if isinstance(key, (Deferred, Builder)):
            key = builder(key)
            if isinstance(key, Variable):
                key = key.name
            else:
                raise TypeError("Only variables can be used as capture keys")
        self.key = key
        self.what = pattern(what)

    def __repr__(self) -> str:
        return f"Capture({self.key!r}, {self.what!r})"

    def equals(self, other: Capture) -> bool:
        return self.key == other.key and self.what == other.what

    @cython.cfunc
    def match(self, value, ctx: Context):
        value = self.what.match(value, ctx)
        ctx[self.key] = value
        return value


@cython.final
@cython.cclass
class Replace(Pattern):
    """Pattern that replaces a value with the output of another pattern.

    Parameters
    ----------
    matcher
        The pattern to match against.
    replacer
        The deferred to use as a replacement.

    """

    searcher: Pattern
    replacer: Builder

    def __init__(self, searcher, replacer):
        self.searcher = pattern(searcher)
        self.replacer = builder(replacer, allow_custom=True)

    def __repr__(self) -> str:
        return f"Replace({self.searcher!r}, {self.replacer!r})"

    @cython.cfunc
    def match(self, value, ctx: Context):
        value = self.searcher.match(value, ctx)
        # use the `_` reserved variable to record the value being replaced
        # in the context, so that it can be used in the replacer pattern
        ctx["_"] = value
        # TODO(kszucs): Builder is not cimported so self.replacer.build() cannot be
        # used, hence using .apply() instead
        return self.replacer.apply(ctx)


def ObjectOf(type_, *args, **kwargs) -> Pattern:
    if isinstance(type_, type):
        if len(type_.__match_args__) < len(args):
            raise ValueError(
                "The type to match has fewer `__match_args__` than the number "
                "of positional arguments in the pattern"
            )
        fields = dict(zip(type_.__match_args__, args))
        fields.update(kwargs)
        if len(fields) == 1:
            return ObjectOf1(type_, **fields)
        elif len(fields) == 2:
            return ObjectOf2(type_, **fields)
        elif len(fields) == 3:
            return ObjectOf3(type_, **fields)
        else:
            return ObjectOfN(type_, **fields)
    else:
        return ObjectOfX(type_, *args, **kwargs)


@cython.cfunc
@cython.inline
def _reconstruct(value: Any, changed: dict[str, Any]):
    # use it with caution because it mutates the changed dict
    name: str
    args: tuple[str] = value.__match_args__
    for name in args:
        if name not in changed:
            changed[name] = getattr(value, name)
    return type(value)(**changed)


@cython.final
@cython.cclass
class ObjectOf1(Pattern):
    type_: Any
    field1: str
    pattern1: Pattern

    def __init__(self, type_: Any, **kwargs):
        assert len(kwargs) == 1
        self.type_ = type_
        ((self.field1, pattern1),) = kwargs.items()
        self.pattern1 = pattern(pattern1)

    def __repr__(self) -> str:
        return f"ObjectOf1({self.type_!r}, {self.field1!r}={self.pattern1!r})"

    def equals(self, other: ObjectOf1) -> bool:
        return (
            self.type_ == other.type_
            and self.field1 == other.field1
            and self.pattern1 == other.pattern1
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.type_):
            raise NoMatchError()

        attr1 = getattr(value, self.field1)
        result1 = self.pattern1.match(attr1, ctx)

        if result1 is not attr1:
            changed: dict = {self.field1: result1}
            return _reconstruct(value, changed)
        else:
            return value


@cython.final
@cython.cclass
class ObjectOf2(Pattern):
    type_: Any
    field1: str
    pattern1: Pattern
    field2: str
    pattern2: Pattern

    def __init__(self, type_: Any, **kwargs):
        assert len(kwargs) == 2
        self.type_ = type_
        (self.field1, pattern1), (self.field2, pattern2) = kwargs.items()
        self.pattern1 = pattern(pattern1)
        self.pattern2 = pattern(pattern2)

    def __repr__(self) -> str:
        return f"ObjectOf2({self.type_!r}, {self.field1!r}={self.pattern1!r}, {self.field2!r}={self.pattern2!r})"

    def equals(self, other: ObjectOf2) -> bool:
        return (
            self.type_ == other.type_
            and self.field1 == other.field1
            and self.pattern1 == other.pattern1
            and self.field2 == other.field2
            and self.pattern2 == other.pattern2
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.type_):
            raise NoMatchError()

        attr1 = getattr(value, self.field1)
        result1 = self.pattern1.match(attr1, ctx)

        attr2 = getattr(value, self.field2)
        result2 = self.pattern2.match(attr2, ctx)

        if result1 is not attr1 or result2 is not attr2:
            changed: dict = {self.field1: result1, self.field2: result2}
            return _reconstruct(value, changed)
        else:
            return value


@cython.final
@cython.cclass
class ObjectOf3(Pattern):
    type_: Any
    field1: str
    field2: str
    field3: str
    pattern1: Pattern
    pattern2: Pattern
    pattern3: Pattern

    def __init__(self, type_: Any, **kwargs):
        assert len(kwargs) == 3
        self.type_ = type_
        (self.field1, pattern1), (self.field2, pattern2), (self.field3, pattern3) = (
            kwargs.items()
        )
        self.pattern1 = pattern(pattern1)
        self.pattern2 = pattern(pattern2)
        self.pattern3 = pattern(pattern3)

    def __repr__(self) -> str:
        return f"ObjectOf3({self.type_!r}, {self.field1!r}={self.pattern1!r}, {self.field2!r}={self.pattern2!r}, {self.field3!r}={self.pattern3!r})"

    def equals(self, other: ObjectOf3) -> bool:
        return (
            self.type_ == other.type_
            and self.field1 == other.field1
            and self.pattern1 == other.pattern1
            and self.field2 == other.field2
            and self.pattern2 == other.pattern2
            and self.field3 == other.field3
            and self.pattern3 == other.pattern3
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.type_):
            raise NoMatchError()

        attr1 = getattr(value, self.field1)
        result1 = self.pattern1.match(attr1, ctx)

        attr2 = getattr(value, self.field2)
        result2 = self.pattern2.match(attr2, ctx)

        attr3 = getattr(value, self.field3)
        result3 = self.pattern3.match(attr3, ctx)

        if result1 is not attr1 or result2 is not attr2 or result3 is not attr3:
            changed: dict = {
                self.field1: result1,
                self.field2: result2,
                self.field3: result3,
            }
            return _reconstruct(value, changed)
        else:
            return value


@cython.final
@cython.cclass
class ObjectOfN(Pattern):
    """Pattern that matches if the object has the given attributes and they match the given patterns.

    The type must conform the structural pattern matching protocol, e.g. it must have a
    __match_args__ attribute that is a tuple of the names of the attributes to match.

    Parameters
    ----------
    type
        The type of the object.
    *args
        The positional arguments to match against the attributes of the object.
    **kwargs
        The keyword arguments to match against the attributes of the object.

    """

    type_: Any
    fields: dict[str, Pattern]

    def __init__(self, type_: Any, **kwargs):
        self.type_ = type_
        self.fields = {k: pattern(v) for k, v in kwargs.items()}

    def __repr__(self) -> str:
        return f"ObjectOfN({self.type_!r}, {self.fields!r})"

    def equals(self, other: ObjectOfN) -> bool:
        return self.type_ == other.type_ and self.fields == other.fields

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, self.type_):
            raise NoMatchError()

        pattern: Pattern
        changed: dict[str, Any] = {}
        for name, pattern in self.fields.items():
            attr = getattr(value, name)
            result = pattern.match(attr, ctx)
            if result is not attr:
                changed[name] = result

        if changed:
            return _reconstruct(value, changed)
        else:
            return value


@cython.final
@cython.cclass
class ObjectOfX(Pattern):
    type_: Pattern
    args: list[Pattern]
    kwargs: dict[str, Pattern]

    def __init__(self, type_, *args, **kwargs):
        self.type_ = pattern(type_)
        self.args = [pattern(arg) for arg in args]
        self.kwargs = {k: pattern(v) for k, v in kwargs.items()}

    def __repr__(self) -> str:
        return f"ObjectOfX({self.type_!r}, {self.args!r}, {self.kwargs!r})"

    def equals(self, other: ObjectOfX) -> bool:
        return (
            self.type_ == self.type_
            and self.args == other.args
            and self.kwargs == other.kwargs
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        self.type_.match(value, ctx)

        # the pattern requirest more positional arguments than the object has
        if len(value.__match_args__) < len(self.args):
            raise NoMatchError()

        patterns: dict[str, Pattern] = dict(zip(value.__match_args__, self.args))
        patterns.update(self.kwargs)

        name: str
        pattern: Pattern
        changed: dict[str, Any] = {}
        for name, pattern in patterns.items():
            try:
                attr = getattr(value, name)
            except AttributeError:
                raise NoMatchError()

            result = pattern.match(attr, ctx)
            if result is not attr:
                changed[name] = result

        if changed:
            return _reconstruct(value, changed)
        else:
            return value


@cython.final
@cython.cclass
class CallableWith(Pattern):
    args: list[Pattern]
    return_: Pattern

    def __init__(self, args, return_=_any):
        self.args = [pattern(arg) for arg in args]
        self.return_ = pattern(return_)

    def __repr__(self) -> str:
        return f"CallableWith({self.args!r}, return_={self.return_!r})"

    def equals(self, other: CallableWith) -> bool:
        return self.args == other.args and self.return_ == other.return_

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not callable(value):
            raise NoMatchError()

        sig = Signature.from_callable(value)

        has_varargs: bool = False
        positional: list = []
        required_positional: list = []
        for p in sig.parameters.values():
            if p.kind in (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD):
                positional.append(p)
                if p.default is Parameter.empty:
                    required_positional.append(p)
            elif p.kind is Parameter.KEYWORD_ONLY and p.default is Parameter.empty:
                raise TypeError(
                    "Callable has mandatory keyword-only arguments which cannot be specified"
                )
            elif p.kind is Parameter.VAR_POSITIONAL:
                has_varargs = True

        if len(required_positional) > len(self.args):
            # Callable has more positional arguments than expected")
            raise NoMatchError()
        elif len(positional) < len(self.args) and not has_varargs:
            # Callable has less positional arguments than expected")
            raise NoMatchError()
        else:
            return value


@cython.final
@cython.cclass
class WithLength(Pattern):
    """Pattern that matches if the length of a value is within a given range.

    Parameters
    ----------
    exactly
        The exact length of the value. If specified, `at_least` and `at_most`
        must be None.
    at_least
        The minimum length of the value.
    at_most
        The maximum length of the value.

    """

    at_least: int
    at_most: int

    def __init__(
        self,
        exactly: Optional[int] = None,
        at_least: Optional[int] = None,
        at_most: Optional[int] = None,
    ):
        if exactly is not None:
            if at_least is not None or at_most is not None:
                raise ValueError("Can't specify both exactly and at_least/at_most")
            at_least = exactly
            at_most = exactly
        self.at_least = at_least
        self.at_most = at_most

    def __repr__(self) -> str:
        return f"WithLength(at_least={self.at_least}, at_most={self.at_most})"

    def equals(self, other: WithLength) -> bool:
        return self.at_least == other.at_least and self.at_most == other.at_most

    @cython.cfunc
    def match(self, value, ctx: Context):
        length = len(value)
        if self.at_least is not None and length < self.at_least:
            raise NoMatchError()
        if self.at_most is not None and length > self.at_most:
            raise NoMatchError()
        return value


def SomeOf(*args, type_=list, **kwargs) -> Pattern:
    if len(args) == 1:
        return SomeItemsOf(*args, type_=type_, **kwargs)
    else:
        return SomeChunksOf(*args, type_=type_, **kwargs)


@cython.final
@cython.cclass
class SomeItemsOf(Pattern):
    pattern: SequenceOf
    delimiter: Pattern
    length: WithLength

    def __init__(self, item, type_=list, **kwargs):
        self.pattern = SequenceOf(item, type_=type_)
        self.delimiter = self.pattern.item
        self.length = WithLength(**kwargs)

    def __repr__(self) -> str:
        return f"SomeItemsOf({self.pattern!r})"

    def equals(self, other: SomeItemsOf) -> bool:
        return self.pattern == other.pattern

    @cython.cfunc
    def match(self, values, ctx: Context):
        result = self.pattern.match(values, ctx)
        return self.length.match(result, ctx)


@cython.final
@cython.cclass
class SomeChunksOf(Pattern):
    """Pattern that unpacks a value into its elements.

    Designed to be used inside a `PatternList` pattern with the `*` syntax.
    """

    pattern: SequenceOf
    delimiter: Pattern
    length: WithLength

    def __init__(self, *args, type_=list, **kwargs):
        pl = PatternList(args)
        self.pattern = SequenceOf(pl, type_=type_)
        self.delimiter = cython.cast(Pattern, pl.delimiter)
        self.length = WithLength(**kwargs)

    def __repr__(self) -> str:
        return f"SomeChunksOf({self.pattern!r}, {self.delimiter!r})"

    def equals(self, other: SomeChunksOf) -> bool:
        return self.pattern == other.pattern and self.delimiter == other.delimiter

    def chunk(self, values, context):
        chunk: list = []
        for item in values:
            try:
                self.delimiter.match(item, context)
            except NoMatchError:
                chunk.append(item)
            else:
                if chunk:  # only yield if there are items in the chunk
                    yield chunk
                chunk = [item]  # start a new chunk with the delimiter
        if chunk:
            yield chunk

    @cython.cfunc
    def match(self, values, ctx: Context):
        chunks = self.chunk(values, ctx)
        result = self.pattern.match(chunks, ctx)
        result = self.length.match(result, ctx)
        return [el for lst in result for el in lst]


def _maybe_unwrap_capture(obj):
    if isinstance(obj, Capture):
        return cython.cast(Capture, obj).what
    else:
        return obj


def PatternList(patterns, type=list):
    if patterns == ():
        return EqValue(patterns)

    patterns = tuple(map(pattern, patterns))
    for pat in patterns:
        pat = _maybe_unwrap_capture(pat)
        if isinstance(pat, (SomeItemsOf, SomeChunksOf)):
            return VariadicPatternList(patterns, type)

    return FixedPatternList(patterns, type)


@cython.final
@cython.cclass
class FixedPatternList(Pattern):
    """Pattern that matches if the respective items in a tuple match the given patterns.

    Parameters
    ----------
    fields
        The patterns to match the respective items in the tuple.

    """

    patterns: list[Pattern]
    type_: Any

    def __init__(self, patterns, type):
        self.patterns = list(map(pattern, patterns))
        self.type_ = type

    def __repr__(self) -> str:
        return f"FixedPatternList({self.patterns!r}, type_={self.type_!r})"

    def equals(self, other: FixedPatternList) -> bool:
        return self.patterns == other.patterns and self.type_ == other.type_

    @property
    def delimiter(self) -> Pattern:
        return self.patterns[0]

    @cython.cfunc
    def match(self, values, ctx: Context):
        if isinstance(values, (str, bytes)):
            raise NoMatchError()

        try:
            values = list(values)
        except TypeError:
            raise NoMatchError()

        if len(values) != len(self.patterns):
            raise NoMatchError()

        result = []
        pattern: Pattern
        for pattern, value in zip(self.patterns, values):
            value = pattern.match(value, ctx)
            result.append(value)

        return self.type_(result)


@cython.final
@cython.cclass
class VariadicPatternList(Pattern):
    patterns: list[Pattern]
    type_: Any

    def __init__(self, patterns, type=list):
        self.patterns = list(map(pattern, patterns))
        self.type_ = type

    def __repr__(self) -> str:
        return f"VariadicPatternList({self.patterns!r}, {self.type_!r})"

    def equals(self, other: VariadicPatternList) -> bool:
        return self.patterns == other.patterns and self.type_ == other.type_

    @property
    def delimiter(self) -> Pattern:
        return self.patterns[0]

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not self.patterns:
            if value:
                raise NoMatchError()
            else:
                return self.type_(value)

        it = RewindableIterator(value)

        result: list = []
        current: Pattern
        original: Pattern
        following: Pattern
        following_patterns = self.patterns[1:] + [Nothing()]
        for current, following in zip(self.patterns, following_patterns):
            original = current
            current = _maybe_unwrap_capture(current)
            following = _maybe_unwrap_capture(following)

            if isinstance(current, (SomeItemsOf, SomeChunksOf)):
                if isinstance(following, SomeItemsOf):
                    following = cython.cast(SomeItemsOf, following).delimiter
                elif isinstance(following, SomeChunksOf):
                    following = cython.cast(SomeChunksOf, following).delimiter

                matches = []
                while True:
                    it.checkpoint()
                    try:
                        item = next(it)
                    except StopIteration:
                        break

                    try:
                        res = following.match(item, ctx)
                    except NoMatchError:
                        matches.append(item)
                    else:
                        it.rewind()
                        break

                res = original.match(matches, ctx)
                result.extend(res)
            else:
                try:
                    item = next(it)
                except StopIteration:
                    raise NoMatchError()

                res = original.match(item, ctx)
                result.append(res)

        return self.type_(result)


@cython.ccall
def PatternMap(fields) -> Pattern:
    if len(fields) == 1:
        return PatternMap1(fields)
    elif len(fields) == 2:
        return PatternMap2(fields)
    else:
        return PatternMapN(fields)


@cython.final
@cython.cclass
class PatternMap1(Pattern):
    field1: str
    pattern1: Pattern

    def __init__(self, fields):
        ((self.field1, pattern1),) = fields.items()
        self.pattern1 = pattern(pattern1)

    def __repr__(self) -> str:
        return f"PatternMap1({self.field1!r}={self.pattern1!r})"

    def equals(self, other: PatternMap1) -> bool:
        return self.field1 == other.field1 and self.pattern1 == other.pattern1

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, dict):
            raise NoMatchError()

        if len(value) != 1:
            raise NoMatchError()

        try:
            item1 = value[self.field1]
        except KeyError:
            raise NoMatchError()

        result1 = self.pattern1.match(item1, ctx)

        if result1 is not item1:
            return type(value)({**value, self.field1: result1})
        else:
            return value


@cython.final
@cython.cclass
class PatternMap2(Pattern):
    field1: str
    field2: str
    pattern1: Pattern
    pattern2: Pattern

    def __init__(self, fields):
        (self.field1, pattern1), (self.field2, pattern2) = fields.items()
        self.pattern1 = pattern(pattern1)
        self.pattern2 = pattern(pattern2)

    def __repr__(self) -> str:
        return f"PatternMap2({self.field1!r}={self.pattern1!r}, {self.field2!r}={self.pattern2!r})"

    def equals(self, other: PatternMap2) -> bool:
        return (
            self.field1 == other.field1
            and self.field2 == other.field2
            and self.pattern1 == other.pattern1
            and self.pattern2 == other.pattern2
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, dict):
            raise NoMatchError()

        if len(value) != 2:
            raise NoMatchError()

        try:
            item1 = value[self.field1]
            item2 = value[self.field2]
        except KeyError:
            raise NoMatchError()

        result1 = self.pattern1.match(item1, ctx)
        result2 = self.pattern2.match(item2, ctx)

        if result1 is not item1 or result2 is not item2:
            return type(value)({**value, self.field1: result1, self.field2: result2})
        else:
            return value


@cython.final
@cython.cclass
class PatternMap3(Pattern):
    field1: str
    field2: str
    field3: str
    pattern1: Pattern
    pattern2: Pattern
    pattern3: Pattern

    def __init__(self, fields):
        (self.field1, pattern1), (self.field2, pattern2), (self.field3, pattern3) = (
            fields.items()
        )
        self.pattern1 = pattern(pattern1)
        self.pattern2 = pattern(pattern2)
        self.pattern3 = pattern(pattern3)

    def __repr__(self) -> str:
        return f"PatternMap3({self.field1!r}={self.pattern1!r}, {self.field2!r}={self.pattern2!r}, {self.field3!r}={self.pattern3!r})"

    def equals(self, other: PatternMap3) -> bool:
        return (
            self.field1 == other.field1
            and self.field2 == other.field2
            and self.field3 == other.field3
            and self.pattern1 == other.pattern1
            and self.pattern2 == other.pattern2
            and self.pattern3 == other.pattern3
        )

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, dict):
            raise NoMatchError()

        if len(value) != 3:
            raise NoMatchError()

        try:
            item1 = value[self.field1]
            item2 = value[self.field2]
            item3 = value[self.field3]
        except KeyError:
            raise NoMatchError()

        result1 = self.pattern1.match(item1, ctx)
        result2 = self.pattern2.match(item2, ctx)
        result3 = self.pattern3.match(item3, ctx)

        if result1 is not item1 or result2 is not item2 or result3 is not item3:
            return type(value)(
                {
                    **value,
                    self.field1: result1,
                    self.field2: result2,
                    self.field3: result3,
                }
            )
        else:
            return value


@cython.final
@cython.cclass
class PatternMapN(Pattern):
    fields: dict[str, Pattern]

    def __init__(self, fields):
        self.fields = {k: pattern(v) for k, v in fields.items()}

    def __repr__(self) -> str:
        return f"PatternMapN({self.fields!r})"

    def equals(self, other: PatternMapN) -> bool:
        return self.fields == other.fields

    @cython.cfunc
    def match(self, value, ctx: Context):
        if not isinstance(value, dict):  # check for __getitem__
            raise NoMatchError()

        if len(value) != len(self.fields):
            raise NoMatchError()

        name: str
        pattern: Pattern
        changed: dict[str, Any] = {}
        for name, pattern in self.fields.items():
            try:
                item = value[name]
            except KeyError:
                raise NoMatchError()
            result = pattern.match(item, ctx)
            if result is not item:
                changed[name] = result

        if changed:
            return type(value)({**value, **changed})
        else:
            return value


@cython.ccall
def pattern(obj: Any, allow_custom: bool = True) -> Pattern:
    """Create a pattern from various types.

    Not that if a Coercible type is passed as argument, the constructed pattern
    won't attempt to coerce the value during matching. In order to allow type
    coercions use `Pattern.from_typehint()` factory method.

    Parameters
    ----------
    obj
        The object to create a pattern from. Can be a pattern, a type, a callable,
        a mapping, an iterable or a value.
    allow_custom
        Whether to allow custom functions to be used as patterns.

    Examples
    --------
    >>> assert pattern(Any()) == Any()
    >>> assert pattern(int) == InstanceOf(int)
    >>>
    >>> @pattern
    ... def as_int(x, context):
    ...     return int(x)
    >>>
    >>> assert as_int.match(1, {}) == 1

    Returns
    -------
    The constructed pattern.

    """
    if obj is Ellipsis:
        return _any
    elif isinstance(obj, Pattern):
        return obj
    elif isinstance(obj, (Deferred, Builder)):
        return EqDeferred(obj)
    elif isinstance(obj, Mapping):
        return PatternMap(obj)
    elif isinstance(obj, Sequence):
        if isinstance(obj, (str, bytes)):
            return EqValue(obj)
        else:
            return PatternList(obj, type(obj))
    elif isinstance(obj, type):
        return InstanceOf(obj)
    elif get_type_origin(obj):
        return Pattern.from_typehint(obj, allow_coercion=False)
    elif callable(obj) and allow_custom:
        return Custom(obj)
    else:
        return EqValue(obj)



from __future__ import annotations

import inspect
from typing import Dict, Generic, List, Optional, TypeVar, Union

import pytest
from typing_extensions import Self

from koerce.utils import (
    get_type_boundvars,
    get_type_hints,
    get_type_params,
)

T = TypeVar("T", covariant=True)
S = TypeVar("S", covariant=True)
U = TypeVar("U", covariant=True)


class My(Generic[T, S, U]):
    a: T
    b: S
    c: str

    @property
    def d(self) -> Optional[str]: ...

    @property
    def e(self) -> U:  # type: ignore
        ...


class MyChild(My): ...


def example(a: int, b: str) -> str:  # type: ignore
    ...


def test_get_type_hints() -> None:
    hints = get_type_hints(My)
    assert hints == {"a": T, "b": S, "c": str}

    hints = get_type_hints(My, include_properties=True)
    assert hints == {"a": T, "b": S, "c": str, "d": Optional[str], "e": U}

    hints = get_type_hints(MyChild, include_properties=True)
    assert hints == {"a": T, "b": S, "c": str, "d": Optional[str], "e": U}

    # test that we don't actually mutate the My.__annotations__
    hints = get_type_hints(My)
    assert hints == {"a": T, "b": S, "c": str}

    hints = get_type_hints(example)
    assert hints == {"a": int, "b": str, "return": str}

    hints = get_type_hints(example, include_properties=True)
    assert hints == {"a": int, "b": str, "return": str}


def test_type_hints_with_fake_type():
    annots = inspect.get_annotations(My)
    hints = get_type_hints(annots, module=__name__)
    assert hints == {"a": T, "b": S, "c": str}

    annots = {"a": "Union[int, str]", "b": "Optional[str]"}
    hints = get_type_hints(annots, module=__name__)
    assert hints == {"a": Union[int, str], "b": Optional[str]}

    annots = {"a": "Union[int, Self]", "b": "Optional[Self]"}
    hints = get_type_hints(annots, module=__name__)
    assert hints == {"a": Union[int, Self], "b": Optional[Self]}


class A(Generic[T, S, U]):
    a: int
    b: str

    t: T
    s: S

    @property
    def u(self) -> U:  # type: ignore
        ...


class B(A[T, S, bytes]): ...


class C(B[T, str]): ...


class D(C[bool]): ...


class MyDict(Dict[T, U]): ...


class MyList(List[T]): ...


def test_get_type_params() -> None:
    # collecting all type params is done by GenericCoercedTo
    assert get_type_params(A[int, float, str]) == {"T": int, "S": float, "U": str}
    assert get_type_params(B[int, bool]) == {"T": int, "S": bool}
    assert get_type_params(C[int]) == {"T": int}
    assert get_type_params(D) == {}
    assert get_type_params(MyDict[int, str]) == {"T": int, "U": str}
    assert get_type_params(MyList[int]) == {"T": int}


def test_get_type_boundvars() -> None:
    expected = {
        "t": int,
        "s": float,
        "u": str,
    }
    assert get_type_boundvars(A[int, float, str]) == expected

    expected = {
        "t": int,
        "s": bool,
    }
    assert get_type_boundvars(B[int, bool]) == expected


def test_get_type_boundvars_unable_to_deduce() -> None:
    msg = "Unable to deduce corresponding type attributes..."
    with pytest.raises(ValueError, match=msg):
        get_type_boundvars(MyDict[int, str])



from __future__ import annotations

from dataclasses import dataclass
from typing import (
    Annotated,
    Optional,
    Union,
)

import pytest

from koerce.annots import (
    EMPTY,
    Parameter,
    Signature,
    ValidationError,
    annotated,
)
from koerce.patterns import (
    Anything,
    InstanceOf,
    MappingOf,
    NoMatchError,
    Option,
    PatternMap,
    TupleOf,
    pattern,
)


def test_parameter():
    p = Parameter("x", Parameter.POSITIONAL_OR_KEYWORD, typehint=int)
    assert p.name == "x"
    assert p.kind is Parameter.POSITIONAL_OR_KEYWORD
    assert str(p) == "x: int"

    p = Parameter("x", Parameter.POSITIONAL_OR_KEYWORD, default=1)
    assert p.name == "x"
    assert p.kind is Parameter.POSITIONAL_OR_KEYWORD
    assert p.default_ == 1
    assert str(p) == "x=1"

    p = Parameter("x", Parameter.POSITIONAL_OR_KEYWORD, typehint=int, default=1)
    assert p.name == "x"
    assert p.kind is Parameter.POSITIONAL_OR_KEYWORD
    assert p.default_ == 1
    assert p.typehint is int
    assert str(p) == "x: int = 1"

    p = Parameter("y", Parameter.VAR_POSITIONAL, typehint=int)
    assert p.name == "y"
    assert p.kind is Parameter.VAR_POSITIONAL
    assert p.typehint is int
    assert str(p) == "*y: int"

    p = Parameter("z", Parameter.VAR_KEYWORD, typehint=int)
    assert p.name == "z"
    assert p.kind is Parameter.VAR_KEYWORD
    assert p.typehint is int
    assert str(p) == "**z: int"


def test_signature_contruction():
    a = Parameter("a", Parameter.POSITIONAL_OR_KEYWORD, typehint=int)
    b = Parameter("b", Parameter.POSITIONAL_OR_KEYWORD, typehint=str)
    c = Parameter("c", Parameter.POSITIONAL_OR_KEYWORD, typehint=int, default=1)
    d = Parameter("d", Parameter.VAR_POSITIONAL, typehint=int)

    sig = Signature([a, b, c, d])
    assert sig.parameters == [a, b, c, d]
    assert sig.return_typehint is EMPTY


def test_signature_from_callable():
    def func(a: int, b: str, *args, c=1, **kwargs) -> float: ...

    sig = Signature.from_callable(func)
    assert sig.parameters == [
        Parameter("a", Parameter.POSITIONAL_OR_KEYWORD, typehint=int),
        Parameter("b", Parameter.POSITIONAL_OR_KEYWORD, typehint=str),
        Parameter("args", Parameter.VAR_POSITIONAL),
        Parameter("c", Parameter.KEYWORD_ONLY, default=1),
        Parameter("kwargs", Parameter.VAR_KEYWORD),
    ]
    assert sig.return_typehint is float


def test_signature_bind_various():
    # with positional or keyword default
    def func(a: int, b: str, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2")
    assert bound == {"a": 1, "b": "2", "c": 1}

    # with variable positional arguments
    def func(a: int, b: str, *args: int, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2", 3, 4)
    assert bound == {"a": 1, "b": "2", "args": (3, 4), "c": 1}

    # with both variadic positional and variadic keyword arguments
    def func(a: int, b: str, *args: int, c=1, **kwargs: int) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2", 3, 4, x=5, y=6)
    assert bound == {
        "a": 1,
        "b": "2",
        "args": (3, 4),
        "c": 1,
        "kwargs": {"x": 5, "y": 6},
    }

    # with positional only arguments
    def func(a: int, b: str, /, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2")
    assert bound == {"a": 1, "b": "2", "c": 1}

    with pytest.raises(TypeError, match="passed as keyword argument"):
        sig(a=1, b="2", c=3)

    # with keyword only arguments
    def func(a: int, b: str, *, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2", c=3)
    assert bound == {"a": 1, "b": "2", "c": 3}

    with pytest.raises(TypeError, match="too many positional arguments"):
        sig(1, "2", 3)

    def func(a, *args, b, z=100, **kwargs): ...

    sig = Signature.from_callable(func)
    bound = sig(10, 20, b=30, c=40, args=50, kwargs=60)
    assert bound == {
        "a": 10,
        "args": (20,),
        "b": 30,
        "z": 100,
        "kwargs": {"c": 40, "args": 50, "kwargs": 60},
    }


def call(func, *args, **kwargs):
    # it also tests the unbind method
    sig = Signature.from_callable(func)
    bound = sig(*args, **kwargs)
    ubargs, ubkwargs = sig.unbind(bound)
    return func(*ubargs, **ubkwargs)


def test_signature_bind_no_arguments():
    def func(): ...

    sig = Signature.from_callable(func)
    assert sig() == {}

    with pytest.raises(TypeError, match="too many positional arguments"):
        sig(1)
    with pytest.raises(TypeError, match="too many positional arguments"):
        sig(1, keyword=2)
    with pytest.raises(TypeError, match="got an unexpected keyword argument 'keyword'"):
        sig(keyword=1)


def test_signature_bind_positional_or_keyword_arguments():
    def func(a, b, c):
        return a, b, c

    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func)
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, 1)
    with pytest.raises(TypeError, match="missing a required argument: 'c'"):
        call(func, 1, 2)
    assert call(func, 1, 2, 3) == (1, 2, 3)

    # one optional argument
    def func(a, b, c=0):
        return a, b, c

    assert call(func, 1, 2, 3) == (1, 2, 3)
    assert call(func, 1, 2) == (1, 2, 0)

    # two optional arguments
    def func(a, b=0, c=0):
        return a, b, c

    assert call(func, 1, 2, 3) == (1, 2, 3)
    assert call(func, 1, 2) == (1, 2, 0)
    assert call(func, 1) == (1, 0, 0)

    # three optional arguments
    def func(a=0, b=0, c=0):
        return a, b, c

    assert call(func, 1, 2, 3) == (1, 2, 3)
    assert call(func, 1, 2) == (1, 2, 0)
    assert call(func, 1) == (1, 0, 0)
    assert call(func) == (0, 0, 0)


def test_signature_bind_varargs():
    def func(*args):
        return args

    assert call(func) == ()
    assert call(func, 1) == (1,)
    assert call(func, 1, 2) == (1, 2)
    assert call(func, 1, 2, 3) == (1, 2, 3)

    def func(a, b, c=3, *args):
        return a, b, c, args

    assert call(func, 1, 2) == (1, 2, 3, ())
    assert call(func, 1, 2, 3) == (1, 2, 3, ())
    assert call(func, 1, 2, 3, 4) == (1, 2, 3, (4,))
    assert call(func, 1, 2, 3, 4, 5) == (1, 2, 3, (4, 5))
    assert call(func, 1, 2, 4) == (1, 2, 4, ())
    assert call(func, a=1, b=2, c=3) == (1, 2, 3, ())
    assert call(func, c=3, a=1, b=2) == (1, 2, 3, ())

    with pytest.raises(TypeError, match="multiple values for argument 'c'"):
        call(func, 1, 2, 3, c=4)

    def func(a, *args):
        return a, args

    with pytest.raises(TypeError, match="got an unexpected keyword argument 'args'"):
        call(func, a=0, args=1)

    def func(*args, **kwargs):
        return args, kwargs

    assert call(func, args=1) == ((), {"args": 1})

    sig = Signature.from_callable(func)
    ba = sig(args=1)
    assert ba == {"args": (), "kwargs": {"args": 1}}


def test_signature_bind_varkwargs():
    def func(**kwargs):
        return kwargs

    assert call(func) == {}
    assert call(func, foo="bar") == {"foo": "bar"}
    assert call(func, foo="bar", spam="ham") == {"foo": "bar", "spam": "ham"}

    def func(a, b, c=3, **kwargs):
        return a, b, c, kwargs

    assert call(func, 1, 2) == (1, 2, 3, {})
    assert call(func, 1, 2, foo="bar") == (1, 2, 3, {"foo": "bar"})
    assert call(func, 1, 2, foo="bar", spam="ham") == (
        1,
        2,
        3,
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, 1, 2, foo="bar", spam="ham", c=4) == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, 1, 2, c=4, foo="bar", spam="ham") == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, 1, 2, c=4, foo="bar", spam="ham", args=10) == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham", "args": 10},
    )
    assert call(func, b=2, a=1, c=4, foo="bar", spam="ham") == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham"},
    )


def test_signature_bind_varargs_and_varkwargs():
    def func(*args, **kwargs):
        return args, kwargs

    assert call(func) == ((), {})
    assert call(func, 1) == ((1,), {})
    assert call(func, 1, 2) == ((1, 2), {})
    assert call(func, foo="bar") == ((), {"foo": "bar"})
    assert call(func, 1, foo="bar") == ((1,), {"foo": "bar"})
    assert call(func, args=10), () == {"args": 10}
    assert call(func, 1, 2, foo="bar") == ((1, 2), {"foo": "bar"})
    assert call(func, 1, 2, foo="bar", spam="ham") == (
        (1, 2),
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, foo="bar", spam="ham", args=10) == (
        (),
        {"foo": "bar", "spam": "ham", "args": 10},
    )


def test_signature_bind_positional_only_arguments():
    def func(a, b, /, c=3):
        return a, b, c

    assert call(func, 1, 2) == (1, 2, 3)
    assert call(func, 1, 2, 4) == (1, 2, 4)
    assert call(func, 1, 2, c=4) == (1, 2, 4)
    with pytest.raises(TypeError, match="multiple values for argument 'c'"):
        call(func, 1, 2, 3, c=4)

    def func(a, b=2, /, c=3, *args):
        return a, b, c, args

    assert call(func, 1, 2) == (1, 2, 3, ())
    assert call(func, 1, 2, 4) == (1, 2, 4, ())
    assert call(func, 1, c=3) == (1, 2, 3, ())

    def func(a, b, c=3, /, foo=42, *, bar=50, **kwargs):
        return a, b, c, foo, bar, kwargs

    assert call(func, 1, 2, 4, 5, bar=6) == (1, 2, 4, 5, 6, {})
    assert call(func, 1, 2) == (1, 2, 3, 42, 50, {})
    assert call(func, 1, 2, foo=4, bar=5) == (1, 2, 3, 4, 5, {})
    assert call(func, 1, 2, foo=4, bar=5, c=10) == (1, 2, 3, 4, 5, {"c": 10})
    assert call(func, 1, 2, 30, c=31, foo=4, bar=5) == (1, 2, 30, 4, 5, {"c": 31})
    assert call(func, 1, 2, 30, foo=4, bar=5, c=31) == (1, 2, 30, 4, 5, {"c": 31})
    assert call(func, 1, 2, c=4) == (1, 2, 3, 42, 50, {"c": 4})
    assert call(func, 1, 2, c=4, foo=5) == (1, 2, 3, 5, 50, {"c": 4})

    with pytest.raises(
        TypeError, match="positional only argument 'a' passed as keyword argument"
    ):
        call(func, a=1, b=2)

    def func(a=1, b=2, /):
        return a, b

    with pytest.raises(TypeError, match="got an unexpected keyword argument 'a'"):
        call(func, a=3, b=4)

    def func(a, /, **kwargs):
        return a, kwargs

    assert call(func, "pos-only", bar="keyword") == ("pos-only", {"bar": "keyword"})


def test_signature_bind_keyword_only_arguments():
    def func(*, a, b, c=3):
        return a, b, c

    with pytest.raises(TypeError, match="too many positional arguments"):
        call(func, 1)

    assert call(func, a=1, b=2) == (1, 2, 3)
    assert call(func, a=1, b=2, c=4) == (1, 2, 4)

    def func(a, *, b, c=3, **kwargs):
        return a, b, c, kwargs

    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, 1)

    assert call(func, 1, b=2) == (1, 2, 3, {})
    assert call(func, 1, b=2, c=4) == (1, 2, 4, {})

    def func(*, a, b, c=3, foo=42, **kwargs):
        return a, b, c, foo, kwargs

    assert call(func, a=1, b=2) == (1, 2, 3, 42, {})
    assert call(func, a=1, b=2, foo=4) == (1, 2, 3, 4, {})
    assert call(func, a=1, b=2, foo=4, bar=5) == (1, 2, 3, 4, {"bar": 5})
    assert call(func, a=1, b=2, foo=4, bar=5, c=10) == (1, 2, 10, 4, {"bar": 5})
    assert call(func, a=1, b=2, foo=4, bar=5, c=10, spam=6) == (
        1,
        2,
        10,
        4,
        {"bar": 5, "spam": 6},
    )

    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func, b=2)
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, a=1)

    def func(a, *, b):
        return a, b

    assert call(func, 1, b=2) == (1, 2)
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, 1)
    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func, b=2)
    with pytest.raises(TypeError, match="got an unexpected keyword argument 'c'"):
        call(func, a=1, b=2, c=3)
    with pytest.raises(TypeError, match="too many positional arguments"):
        call(func, 1, 2)
    with pytest.raises(TypeError, match="too many positional arguments"):
        call(func, 1, 2, c=3)

    def func(a, *, b, **kwargs):
        return a, b, kwargs

    assert call(func, 1, b=2) == (1, 2, {})
    assert call(func, 1, b=2, c=3) == (1, 2, {"c": 3})
    assert call(func, 1, b=2, c=3, d=4) == (1, 2, {"c": 3, "d": 4})
    assert call(func, a=1, b=2) == (1, 2, {})
    assert call(func, c=3, a=1, b=2) == (1, 2, {"c": 3})
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, a=1)
    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func, c=3, b=2)


def test_signature_bind_with_arg_named_self():
    def test(a, self, b):
        pass

    sig = Signature.from_callable(test)
    ba = sig(1, 2, 3)
    args, _ = sig.unbind(ba)
    assert args == (1, 2, 3)
    ba = sig(1, self=2, b=3)
    args, _ = sig.unbind(ba)
    assert args == (1, 2, 3)


def test_signature_unbind_from_callable():
    def test(a: int, b: int, c: int = 1): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3)

    assert bound == {"a": 2, "b": 3, "c": 1}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 1)
    assert kwargs == {}


def test_signature_unbind_from_callable_with_varargs():
    def test(a: int, b: int, *args: int): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3)

    assert bound == {"a": 2, "b": 3, "args": ()}
    args, kwargs = sig.unbind(bound)
    assert args == (2, 3)
    assert kwargs == {}

    bound = sig(2, 3, 4, 5)
    assert bound == {"a": 2, "b": 3, "args": (4, 5)}
    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 4, 5)
    assert kwargs == {}


def test_signature_unbind_from_callable_with_positional_only_arguments():
    def test(a: int, b: int, /, c: int = 1): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3)
    assert bound == {"a": 2, "b": 3, "c": 1}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 1)
    assert kwargs == {}

    bound = sig(2, 3, 4)
    assert bound == {"a": 2, "b": 3, "c": 4}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 4)
    assert kwargs == {}


def test_signature_unbind_from_callable_with_keyword_only_arguments():
    def test(a: int, b: int, *, c: float, d: float = 0.0): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3, c=4.0)
    assert bound == {"a": 2, "b": 3, "c": 4.0, "d": 0.0}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3)
    assert kwargs == {"c": 4.0, "d": 0.0}


def test_signature_unbind():
    def func(a, b, c=1): ...

    sig = Signature.from_callable(func)
    bound = sig(1, 2)

    assert bound == {"a": 1, "b": 2, "c": 1}

    args, kwargs = sig.unbind(bound)
    assert args == (1, 2, 1)
    assert kwargs == {}


@pytest.mark.parametrize("d", [(), (5, 6, 7)])
def test_signature_unbind_with_empty_variadic(d):
    def func(a, b, c, *args, e=None):
        return a, b, c, args, e

    sig = Signature.from_callable(func)
    bound = sig(1, 2, 3, *d, e=4)
    assert bound == {"a": 1, "b": 2, "c": 3, "args": d, "e": 4}

    args, kwargs = sig.unbind(bound)
    assert args == (1, 2, 3, *d)
    assert kwargs == {"e": 4}


def test_signature_merge():
    def f1(a, b, c=1): ...

    def f2(d, e, f=2): ...

    def f3(d, a=1, **kwargs): ...

    sig1 = Signature.from_callable(f1)
    sig2 = Signature.from_callable(f2)
    sig3 = Signature.from_callable(f3)

    sig12 = Signature.merge(sig1, sig2)
    assert sig12.parameters == [
        Parameter("a", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("b", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("d", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("e", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("c", Parameter.POSITIONAL_OR_KEYWORD, default=1),
        Parameter("f", Parameter.POSITIONAL_OR_KEYWORD, default=2),
    ]

    sig13 = Signature.merge(sig1, sig3)
    assert sig13.parameters == [
        Parameter("b", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("d", Parameter.POSITIONAL_OR_KEYWORD),
        Parameter("a", Parameter.POSITIONAL_OR_KEYWORD, default=1),
        Parameter("c", Parameter.POSITIONAL_OR_KEYWORD, default=1),
        Parameter("kwargs", Parameter.VAR_KEYWORD),
    ]


def test_annotated_function():
    @annotated(a=InstanceOf(int), b=InstanceOf(int), c=InstanceOf(int))
    def test(a, b, c=1):
        return a + b + c

    assert test(2, 3) == 6
    assert test(2, 3, 4) == 9
    assert test(2, 3, c=4) == 9
    assert test(a=2, b=3, c=4) == 9

    with pytest.raises(ValidationError):
        test(2, 3, c="4")

    @annotated(a=InstanceOf(int))
    def test(a, b, c=1):
        return (a, b, c)

    assert test(2, "3") == (2, "3", 1)


def test_annotated_function_with_type_annotations():
    @annotated()
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    assert test(2, 3) == 6

    @annotated
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    assert test(2, 3) == 6

    @annotated
    def test(a: int, b, c=1):
        return (a, b, c)

    assert test(2, 3, "4") == (2, 3, "4")


def test_annotated_function_with_return_type_annotation():
    @annotated
    def test_ok(a: int, b: int, c: int = 1) -> int:
        return a + b + c

    @annotated
    def test_wrong(a: int, b: int, c: int = 1) -> int:
        return "invalid result"

    assert test_ok(2, 3) == 6
    with pytest.raises(ValidationError):
        test_wrong(2, 3)


def test_annotated_function_with_keyword_overrides():
    @annotated(b=InstanceOf(float))
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(ValidationError):
        test(2, 3)

    assert test(2, 3.0) == 6.0


def test_annotated_function_with_list_overrides():
    @annotated([InstanceOf(int), InstanceOf(int), InstanceOf(str)])
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(ValidationError):
        test(2, 3, 4)


def test_annotated_function_with_list_overrides_and_return_override():
    @annotated([InstanceOf(int), InstanceOf(int), InstanceOf(float)], InstanceOf(float))
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(ValidationError):
        test(2, 3, 4)

    assert test(2, 3, 4.0) == 9.0


@pattern
def short_str(x, **context):
    if len(x) > 3:
        return x
    else:
        raise NoMatchError()


@pattern
def endswith_d(x, **context):
    if x.endswith("d"):
        return x
    else:
        raise NoMatchError()


def test_annotated_function_with_complex_type_annotations():
    @annotated
    def test(a: Annotated[str, short_str, endswith_d], b: Union[int, float]):
        return a, b

    assert test("abcd", 1) == ("abcd", 1)
    assert test("---d", 1.0) == ("---d", 1.0)

    with pytest.raises(ValidationError):
        test("---c", 1)
    with pytest.raises(ValidationError):
        test("123", 1)
    with pytest.raises(ValidationError):
        test("abcd", "qweqwe")


def test_annotated_function_without_annotations():
    @annotated
    def test(a, b, c):
        return a, b, c

    assert test(1, 2, 3) == (1, 2, 3)
    assert [p.name for p in test.__signature__.parameters] == ["a", "b", "c"]


def test_annotated_function_without_decoration():
    def test(a, b, c):
        return a + b + c

    func = annotated(test)
    with pytest.raises(TypeError):
        func(1, 2)

    assert func(1, 2, c=3) == 6


def test_annotated_function_with_varargs():
    @annotated
    def test(a: float, b: float, *args: int):
        return sum((a, b) + args)

    assert test(1.0, 2.0, 3, 4) == 10.0
    assert test(1.0, 2.0, 3, 4, 5) == 15.0

    with pytest.raises(ValidationError):
        test(1.0, 2.0, 3, 4, 5, 6.0)


def test_annotated_function_with_varkwargs():
    @annotated
    def test(a: float, b: float, **kwargs: int):
        return sum((a, b) + tuple(kwargs.values()))

    assert test(1.0, 2.0, c=3, d=4) == 10.0
    assert test(1.0, 2.0, c=3, d=4, e=5) == 15.0

    with pytest.raises(ValidationError):
        test(1.0, 2.0, c=3, d=4, e=5, f=6.0)


# def test_multiple_validation_failures():
#     @annotated
#     def test(a: float, b: float, *args: int, **kwargs: int): ...

#     with pytest.raises(ValidationError) as excinfo:
#         test(1.0, 2.0, 3.0, 4, c=5.0, d=6)

#     assert len(excinfo.value.errors) == 2


def test_signature_to_pattern():
    def func(a: int, b: str) -> str: ...

    args, ret = Signature.from_callable(func).to_pattern()
    assert args == PatternMap({"a": InstanceOf(int), "b": InstanceOf(str)})
    assert ret == InstanceOf(str)

    def func(a: int, b: str, c: str = "0") -> str: ...

    args, ret = Signature.from_callable(func).to_pattern()
    assert args == PatternMap(
        {"a": InstanceOf(int), "b": InstanceOf(str), "c": Option(InstanceOf(str), "0")}
    )
    assert ret == InstanceOf(str)

    def func(a: int, b: str, *args): ...

    args, ret = Signature.from_callable(func).to_pattern()
    assert args == PatternMap(
        {"a": InstanceOf(int), "b": InstanceOf(str), "args": TupleOf(Anything())}
    )
    assert ret == Anything()

    def func(a: int, b: str, c: str = "0", *args, **kwargs: int) -> float: ...

    args, ret = Signature.from_callable(func).to_pattern()
    assert args == PatternMap(
        {
            "a": InstanceOf(int),
            "b": InstanceOf(str),
            "c": Option(InstanceOf(str), "0"),
            "args": TupleOf(Anything()),
            "kwargs": MappingOf(Anything(), InstanceOf(int)),
        }
    )
    assert ret == InstanceOf(float)


def test_annotated_with_class():
    @annotated
    class A:
        a: int
        b: str
        c: float
        d: Optional[int]

        def __init__(self, a, b, c, d=1):
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    with pytest.raises(ValidationError):
        A(1, "2", "d")


def test_annotated_with_dataclass():
    @annotated
    @dataclass
    class InventoryItem:
        name: str
        unit_price: float
        quantity_on_hand: int = 0

    item = InventoryItem("widget", 3.0, 10)
    assert item.name == "widget"
    assert item.unit_price == 3.0
    assert item.quantity_on_hand == 10

    item = InventoryItem("widget", 3.0)
    assert item.name == "widget"
    assert item.unit_price == 3.0
    assert item.quantity_on_hand == 0

    with pytest.raises(ValidationError):
        InventoryItem("widget", "3.0", 10)

    with pytest.raises(ValidationError):
        InventoryItem("widget", 3.0, "10")



from __future__ import annotations

import functools
import inspect
import typing
from typing import Any

import cython

from .patterns import (
    DictOf,
    NoMatch,
    Option,
    Pattern,
    PatternMap,
    TupleOf,
    _any,
    pattern,
)
from .utils import get_type_hints

EMPTY = inspect.Parameter.empty
_POSITIONAL_ONLY = cython.declare(cython.int, int(inspect.Parameter.POSITIONAL_ONLY))
_POSITIONAL_OR_KEYWORD = cython.declare(
    cython.int, int(inspect.Parameter.POSITIONAL_OR_KEYWORD)
)
_VAR_POSITIONAL = cython.declare(cython.int, int(inspect.Parameter.VAR_POSITIONAL))
_KEYWORD_ONLY = cython.declare(cython.int, int(inspect.Parameter.KEYWORD_ONLY))
_VAR_KEYWORD = cython.declare(cython.int, int(inspect.Parameter.VAR_KEYWORD))


@cython.final
@cython.cclass
class Parameter:
    POSITIONAL_ONLY: typing.ClassVar[int] = _POSITIONAL_ONLY
    POSITIONAL_OR_KEYWORD: typing.ClassVar[int] = _POSITIONAL_OR_KEYWORD
    VAR_POSITIONAL: typing.ClassVar[int] = _VAR_POSITIONAL
    KEYWORD_ONLY: typing.ClassVar[int] = _KEYWORD_ONLY
    VAR_KEYWORD: typing.ClassVar[int] = _VAR_KEYWORD

    name = cython.declare(str, visibility="readonly")
    kind = cython.declare(cython.int, visibility="readonly")
    # Cannot use C reserved keyword 'default' here
    default_ = cython.declare(object, visibility="readonly")
    typehint = cython.declare(object, visibility="readonly")

    def __init__(
        self, name: str, kind: int, default: Any = EMPTY, typehint: Any = EMPTY
    ):
        self.name = name
        self.kind = kind
        self.default_ = default
        self.typehint = typehint

    def __str__(self) -> str:
        result: str = self.name
        if self.typehint is not EMPTY:
            if hasattr(self.typehint, "__qualname__"):
                result += f": {self.typehint.__qualname__}"
            else:
                result += f": {self.typehint}"
        if self.default_ is not EMPTY:
            if self.typehint is EMPTY:
                result = f"{result}={self.default_}"
            else:
                result = f"{result} = {self.default_!r}"
        if self.kind == _VAR_POSITIONAL:
            result = f"*{result}"
        elif self.kind == _VAR_KEYWORD:
            result = f"**{result}"
        return result

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self}">'

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        right: Parameter = cython.cast(Parameter, other)
        return (
            self.name == right.name
            and self.kind == right.kind
            and self.default_ == right.default_
            and self.typehint == right.typehint
        )


@cython.final
@cython.cclass
class Signature:
    parameters = cython.declare(list[Parameter], visibility="readonly")
    return_typehint = cython.declare(object, visibility="readonly")

    def __init__(self, parameters: list[Parameter], return_typehint: Any = EMPTY):
        self.parameters = parameters
        self.return_typehint = return_typehint

    @staticmethod
    def from_callable(func: Any) -> Signature:
        sig = inspect.signature(func)
        hints = get_type_hints(func)
        params: list[Parameter] = [
            Parameter(p.name, int(p.kind), p.default, hints.get(p.name, EMPTY))
            for p in sig.parameters.values()
        ]
        return Signature(params, return_typehint=hints.get("return", EMPTY))

    @staticmethod
    def merge(*signatures: Signature, **annotations):
        """Merge multiple signatures.

        In addition to concatenating the parameters, it also reorders the
        parameters so that optional arguments come after mandatory arguments.

        Parameters
        ----------
        *signatures : Signature
            Signature instances to merge.
        **annotations : dict
            Annotations to add to the merged signature.

        Returns
        -------
        Signature

        """
        name: str
        param: Parameter
        params: dict[str, Parameter] = {}
        for sig in signatures:
            for param in sig.parameters:
                params[param.name] = param

        inherited: set[str] = set(params.keys())
        for name, annot in annotations.items():
            params[name] = Parameter(name, annotation=annot)

        # mandatory fields without default values must precede the optional
        # ones in the function signature, the partial ordering will be kept
        old_args: list = []
        new_args: list = []
        var_args: list = []
        var_kwargs: list = []
        new_kwargs: list = []
        old_kwargs: list = []
        for name, param in params.items():
            if param.kind == _VAR_POSITIONAL:
                if var_args:
                    raise TypeError("only one variadic *args parameter is allowed")
                var_args.append(param)
            elif param.kind == _VAR_KEYWORD:
                if var_kwargs:
                    raise TypeError("only one variadic **kwargs parameter is allowed")
                var_kwargs.append(param)
            elif name in inherited:
                if param.default_ is EMPTY:
                    old_args.append(param)
                else:
                    old_kwargs.append(param)
            elif param.default_ is EMPTY:
                new_args.append(param)
            else:
                new_kwargs.append(param)

        return Signature(
            old_args + new_args + var_args + new_kwargs + old_kwargs + var_kwargs
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Signature):
            return NotImplemented
        right: Signature = cython.cast(Signature, other)
        return (
            self.parameters == right.parameters
            and self.return_annotation == right.return_annotation
        )

    def __call__(self, /, *args, **kwargs):
        return self.bind(args, kwargs)

    @cython.ccall
    def bind(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> dict[str, Any]:
        """Bind the arguments to the signature.

        Parameters
        ----------
        args : Any
            Positional arguments.
        kwargs : Any
            Keyword arguments.

        Returns
        -------
        dict
            Mapping of parameter names to argument values.

        """
        i: cython.int = 0
        kind: cython.int
        param: Parameter
        bound: dict[str, Any] = {}

        # 1. HANDLE ARGS
        for i in range(len(args)):
            if i >= len(self.parameters):
                raise TypeError("too many positional arguments")

            param = self.parameters[i]
            kind = param.kind
            if kind is _POSITIONAL_OR_KEYWORD:
                if param.name in kwargs:
                    raise TypeError(f"multiple values for argument '{param.name}'")
                bound[param.name] = args[i]
            elif kind is _VAR_KEYWORD or kind is _KEYWORD_ONLY:
                raise TypeError("too many positional arguments")
            elif kind is _VAR_POSITIONAL:
                bound[param.name] = args[i:]
                break
            elif kind is _POSITIONAL_ONLY:
                bound[param.name] = args[i]
            else:
                raise TypeError("unreachable code")

        # 2. INCREMENT PARAMETER INDEX
        if args:
            i += 1

        # 3. HANDLE KWARGS
        for param in self.parameters[i:]:
            if param.kind is _POSITIONAL_OR_KEYWORD or param.kind is _KEYWORD_ONLY:
                if param.name in kwargs:
                    bound[param.name] = kwargs.pop(param.name)
                elif param.default_ is EMPTY:
                    raise TypeError(f"missing a required argument: '{param.name}'")
                else:
                    bound[param.name] = param.default_
            elif param.kind is _VAR_POSITIONAL:
                bound[param.name] = ()
            elif param.kind is _VAR_KEYWORD:
                bound[param.name] = kwargs
                break
            elif param.kind is _POSITIONAL_ONLY:
                if param.default_ is EMPTY:
                    if param.name in kwargs:
                        raise TypeError(
                            f"positional only argument '{param.name}' passed as keyword argument"
                        )
                    else:
                        raise TypeError(
                            f"missing required positional argument {param.name}"
                        )
                else:
                    bound[param.name] = param.default_
            else:
                raise TypeError("unreachable code")
        else:
            if kwargs:
                raise TypeError(
                    f"got an unexpected keyword argument '{next(iter(kwargs))}'"
                )

        return bound

    @cython.ccall
    def unbind(self, bound: dict[str, Any]) -> tuple[tuple[Any, ...], dict[str, Any]]:
        """Reverse bind of the parameters.

        Attempts to reconstructs the original arguments as keyword only arguments.

        Parameters
        ----------
        bound
            Object with attributes matching the signature parameters.

        Returns
        -------
        args : (args, kwargs)
            Tuple of positional and keyword arguments.

        """
        # does the reverse of bind, but doesn't apply defaults
        args: list = []
        kwargs: dict = {}
        param: Parameter
        for param in self.parameters:
            value = bound[param.name]
            if param.kind is _POSITIONAL_OR_KEYWORD:
                args.append(value)
            elif param.kind is _VAR_POSITIONAL:
                args.extend(value)
            elif param.kind is _VAR_KEYWORD:
                kwargs.update(value)
            elif param.kind is _KEYWORD_ONLY:
                kwargs[param.name] = value
            elif param.kind is _POSITIONAL_ONLY:
                args.append(value)
            else:
                raise TypeError(f"unsupported parameter kind {param.kind}")

        return tuple(args), kwargs

    def to_pattern(
        self,
        overrides: dict[str, Any] | list[Any] | None = None,
        return_override: Any = None,
    ) -> Pattern:
        """Create patterns from a Signature.

        Two patterns are created, one for the arguments and one for the return value.

        Parameters
        ----------
        overrides : dict, default None
            Pass patterns to add missing or override existing argument type
            annotations.
        return_override : Pattern, default None
            Pattern for the return value of the callable.

        Returns
        -------
        Tuple of patterns for the arguments and the return value.
        """
        arg_overrides: dict[str, Any]
        if overrides is None:
            arg_overrides = {}
        elif isinstance(overrides, (list, tuple)):
            # create a mapping of parameter name to pattern
            arg_overrides = {
                param.name: arg for param, arg in zip(self.parameters, overrides)
            }
        elif isinstance(overrides, dict):
            arg_overrides = overrides
        else:
            raise TypeError(f"patterns must be a list or dict, got {type(overrides)}")

        retpat: Pattern
        argpat: Pattern
        argpats: dict[str, Pattern] = {}
        for param in self.parameters:
            name: str = param.name

            if name in arg_overrides:
                argpat = pattern(arg_overrides[name])
            elif param.typehint is not EMPTY:
                argpat = Pattern.from_typehint(param.typehint)
            else:
                argpat = _any

            if param.kind is _VAR_POSITIONAL:
                argpat = TupleOf(argpat)
            elif param.kind is _VAR_KEYWORD:
                argpat = DictOf(_any, argpat)
            elif param.default_ is not EMPTY:
                argpat = Option(argpat, default=param.default_)

            argpats[name] = argpat

        if return_override is not None:
            retpat = pattern(return_override)
        elif self.return_typehint is not EMPTY:
            retpat = Pattern.from_typehint(self.return_typehint)
        else:
            retpat = _any

        return (PatternMap(argpats), retpat)


class ValidationError(Exception):
    pass


def annotated(_1=None, _2=None, _3=None, **kwargs):
    """Create functions with arguments validated at runtime.

    There are various ways to apply this decorator:

    1. With type annotations

    >>> @annotated
    ... def foo(x: int, y: str) -> float:
    ...     return float(x) + float(y)

    2. With argument patterns passed as keyword arguments

    >>> from ibis.common.patterns import InstanceOf as instance_of
    >>> @annotated(x=instance_of(int), y=instance_of(str))
    ... def foo(x, y):
    ...     return float(x) + float(y)

    3. With mixing type annotations and patterns where the latter takes precedence

    >>> @annotated(x=instance_of(float))
    ... def foo(x: int, y: str) -> float:
    ...     return float(x) + float(y)

    4. With argument patterns passed as a list and/or an optional return pattern

    >>> @annotated([instance_of(int), instance_of(str)], instance_of(float))
    ... def foo(x, y):
    ...     return float(x) + float(y)

    Parameters
    ----------
    *args : Union[
                tuple[Callable],
                tuple[list[Pattern], Callable],
                tuple[list[Pattern], Pattern, Callable]
            ]
        Positional arguments.
        - If a single callable is passed, it's wrapped with the signature
        - If two arguments are passed, the first one is a list of patterns for the
          arguments and the second one is the callable to wrap
        - If three arguments are passed, the first one is a list of patterns for the
          arguments, the second one is a pattern for the return value and the third
          one is the callable to wrap
    **kwargs : dict[str, Pattern]
        Patterns for the arguments.

    Returns
    -------
    Callable

    """
    if _1 is None:
        return functools.partial(annotated, **kwargs)
    elif _2 is None:
        if callable(_1):
            func, patterns, return_pattern = _1, None, None
        else:
            return functools.partial(annotated, _1, **kwargs)
    elif _3 is None:
        if not isinstance(_2, Pattern):
            func, patterns, return_pattern = _2, _1, None
        else:
            return functools.partial(annotated, _1, _2, **kwargs)
    else:
        func, patterns, return_pattern = _3, _1, _2

    sig: Signature = Signature.from_callable(func)
    argpats, retpat = sig.to_pattern(
        overrides=patterns or kwargs, return_override=return_pattern
    )

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        # 0. Bind the arguments to the signature
        bound: dict[str, Any] = sig.bind(args, kwargs)

        # 1. Validate the passed arguments
        values: Any = argpats.apply(bound)
        if values is NoMatch:
            raise ValidationError()

        # 2. Reconstruction of the original arguments
        args, kwargs = sig.unbind(values)

        # 3. Call the function with the validated arguments
        result = func(*args, **kwargs)

        # 4. Validate the return value
        result = retpat.apply(result)
        if result is NoMatch:
            raise ValidationError()

        return result

    wrapped.__signature__ = sig

    return wrapped



from __future__ import annotations

import itertools
import sys
import typing
from typing import Any, Optional, TypeVar

get_type_args = typing.get_args
get_type_origin = typing.get_origin


class FakeType:
    def __init__(self, module: str, annotations: dict[str, Any]):
        self.__module__ = module
        self.__annotations__ = annotations


def get_type_hints(
    obj: Any,
    globalns=None,
    localns=None,
    module: Optional[str] = None,
    include_extras: bool = True,
    include_properties: bool = False,
) -> dict[str, Any]:
    """Get type hints for a callable or class.

    Extension of typing.get_type_hints that supports getting type hints for
    class properties.

    Parameters
    ----------
    obj
        Callable or class to get type hints for.
    globalns
        Global namespace to use for resolving type variables.
    localns
        Local namespace to use for resolving type variables.
    module
        Module name to use for resolving globalns if not provided.
    include_extras
        Whether to include extra type hints such as Annotated.
    include_properties
        Whether to include type hints for class properties.

    Returns
    -------
    Mapping of parameter or attribute name to type hint.
    """
    if isinstance(obj, dict):
        assert module is not None
        obj = FakeType(module=module, annotations=obj)
        mod = sys.modules.get(module, None)
        globalns = getattr(mod, "__dict__", None)

    try:
        hints = typing.get_type_hints(
            obj, globalns=globalns, localns=localns, include_extras=include_extras
        )
    except TypeError:
        return {}

    if include_properties:
        for name in dir(obj):
            attr = getattr(obj, name)
            if isinstance(attr, property):
                annots = typing.get_type_hints(
                    attr.fget,
                    globalns=globalns,
                    localns=localns,
                    include_extras=include_extras,
                )
                if return_hint := annots.get("return"):
                    hints[name] = return_hint

    return hints


# TODO(kszucs): memoize this function
def get_type_params(typ: Any) -> dict[TypeVar, type]:
    """Get type parameters for a generic class.

    Parameters
    ----------
    typ
        Generic class to get type parameters for.

    Returns
    -------
    Mapping of type parameter name to type.

    Examples
    --------
    >>> from typing import Dict, List
    >>> class MyList(List[T]): ...
    >>> get_type_params(MyList[int])
    {'T': <class 'int'>}
    >>> class MyDict(Dict[T, U]): ...
    >>> get_type_params(MyDict[int, str])
    {'T': <class 'int'>, 'U': <class 'str'>}
    """
    args: tuple = get_type_args(typ)
    origin: Any = get_type_origin(typ)  # or typ
    params: tuple = getattr(origin, "__parameters__", ())

    result: dict[str, type] = {}
    for param, arg in zip(params, args):
        result[param.__name__] = arg

    return result


def get_type_boundvars(typ: Any) -> dict[TypeVar, tuple[str, type]]:
    """Get the attribute names and concrete types of a generic class.

    Parameters
    ----------
    typ
        Generic class to get type variables for.
    only_covariant
        Whether to only include covariant type variables.

    Returns
    -------
    Mapping of type variable to attribute name and type.

    Examples
    --------
    >>> from typing import Generic
    >>> class MyStruct(Generic[T, U]):
    ...     a: T
    ...     b: U
    >>> get_bound_typevars(MyStruct[int, str])
    {~T: ('a', <class 'int'>), ~U: ('b', <class 'str'>)}
    >>>
    >>> class MyStruct(Generic[T, U]):
    ...     a: T
    ...
    ...     @property
    ...     def myprop(self) -> U: ...
    >>> get_bound_typevars(MyStruct[float, bytes])
    {~T: ('a', <class 'float'>), ~U: ('myprop', <class 'bytes'>)}
    """
    origin = get_type_origin(typ) or typ
    hints = get_type_hints(origin, include_properties=True)
    params = get_type_params(typ)

    result = {}
    for attr, hint in hints.items():
        if isinstance(hint, TypeVar) and hint.__name__ in params:
            if not hint.__covariant__:
                raise TypeError(
                    f"Typevar {hint} is not covariant and currently only "
                    "covariant typevars are supported"
                )
            result[attr] = params.pop(hint.__name__)

    if params:
        raise ValueError(
            f"Unable to deduce corresponding type attributes for the following type variables: {params}"
        )

    return result


class RewindableIterator:
    """Iterator that can be rewound to a checkpoint.

    Examples
    --------
    >>> it = RewindableIterator(range(5))
    >>> next(it)
    0
    >>> next(it)
    1
    >>> it.checkpoint()
    >>> next(it)
    2
    >>> next(it)
    3
    >>> it.rewind()
    >>> next(it)
    2
    >>> next(it)
    3
    >>> next(it)
    4

    """

    __slots__ = ("_iterator", "_checkpoint")

    def __init__(self, iterable):
        self._iterator = iter(iterable)
        self._checkpoint = None

    def __next__(self):
        return next(self._iterator)

    def rewind(self):
        """Rewind the iterator to the last checkpoint."""
        if self._checkpoint is None:
            raise ValueError("No checkpoint to rewind to.")
        self._iterator, self._checkpoint = itertools.tee(self._checkpoint)

    def checkpoint(self):
        """Create a checkpoint of the current iterator state."""
        self._iterator, self._checkpoint = itertools.tee(self._iterator)



from __future__ import annotations

import operator

import pytest

from koerce.builders import (
    Attr,
    Binop,
    Call,
    Call0,
    Call1,
    Call2,
    Call3,
    CallN,
    Custom,
    Deferred,
    Item,
    Just,
    Mapping,
    Sequence,
    Unop,
    Variable,
    builder,
)

_ = Deferred(Variable("_"))


def test_builder():
    class MyClass:
        pass

    def fn(x):
        return x + 1

    assert builder(1) == Just(1)
    assert builder(Just(1)) == Just(1)
    assert builder(Just(Just(1))) == Just(1)
    assert builder(MyClass) == Just(MyClass)
    assert builder(fn) == Just(fn)
    assert builder(()) == Sequence(())
    assert builder((1, 2, _)) == Sequence((Just(1), Just(2), _))
    assert builder({}) == Mapping({})
    assert builder({"a": 1, "b": _}) == Mapping({"a": Just(1), "b": _})
    assert builder("string") == Just("string")

    # assert builder(var("x")) == Variable("x")
    # assert builder(Variable("x")) == Variable("x")


def test_builder_just():
    p = Just(1)
    assert p.apply({}) == 1
    assert p.apply({"a": 1}) == 1

    # unwrap subsequently nested Just instances
    assert Just(p) == p

    # disallow creating a Just builder from other builders or deferreds
    # with pytest.raises(TypeError, match="cannot be used as a Just value"):
    #     Just(_)
    # with pytest.raises(TypeError, match="cannot be used as a Just value"):
    #     Just(Factory(lambda _: _))


def test_builder_variable():
    p = Variable("other")
    context = {"other": 10}
    assert p.apply(context) == 10


def test_builder_custom():
    f = Custom(lambda _: _ + 1)
    assert f.apply({"_": 1}) == 2
    assert f.apply({"_": 2}) == 3

    def fn(**kwargs):
        assert kwargs == {"_": 10, "a": 5}
        return -1

    f = Custom(fn)
    assert f.apply({"_": 10, "a": 5}) == -1


def test_builder_call():
    def fn(*args):
        return args

    def func(a, b, c=1):
        return a + b + c

    c = Call0(Just(fn))
    assert c.apply({}) == ()

    c = Call1(Just(fn), Just(1))
    assert c.apply({}) == (1,)

    c = Call2(Just(fn), Just(1), Just(2))
    assert c.apply({}) == (1, 2)

    c = Call3(Just(fn), Just(1), Just(2), Just(3))
    assert c.apply({}) == (1, 2, 3)

    c = Call(Just(func), Just(1), Just(2), c=Just(3))
    assert isinstance(c, CallN)
    assert c.apply({}) == 6

    c = Call(Just(func), Just(-1), Just(-2))
    assert isinstance(c, Call2)
    assert c.apply({}) == -2

    c = Call(Just(dict), a=Just(1), b=Just(2))
    assert isinstance(c, CallN)
    assert c.apply({}) == {"a": 1, "b": 2}

    c = Call(Just(float), Just("1.1"))
    assert isinstance(c, Call1)
    assert c.apply({}) == 1.1

    c = Call(Just(list))
    assert isinstance(c, Call0)
    assert c.apply({}) == []


def test_builder_unop():
    b = Unop(operator.neg, Just(1))
    assert b.apply({}) == -1

    b = Unop(operator.abs, Just(-1))
    assert b.apply({}) == 1


def test_builder_binop():
    b = Binop(operator.add, Just(1), Just(2))
    assert b.apply({}) == 3

    b = Binop(operator.mul, Just(2), Just(3))
    assert b.apply({}) == 6


def test_builder_attr():
    class MyType:
        def __init__(self, a, b):
            self.a = a
            self.b = b

        def __hash__(self):
            return hash((type(self), self.a, self.b))

    v = Variable("v")
    b = Attr(v, "b")
    assert b.apply({"v": MyType(1, 2)}) == 2

    b = Attr(Just(MyType(1, 2)), "a")
    assert b.apply({}) == 1

    # name is not allowed to be a deferred
    b = Attr(v, "a")
    assert b.apply({"v": MyType(1, 2)}) == 1


def test_builder_item():
    v = Variable("v")
    b = Item(v, Just(1))
    assert b.apply({"v": [1, 2, 3]}) == 2

    b = Item(Just(dict(a=1, b=2)), Just("a"))
    assert b.apply({}) == 1

    name = Variable("name")
    # test that name can be a deferred as well
    b = Item(v, name)
    assert b.apply({"v": {"a": 1, "b": 2}, "name": "b"}) == 2


def test_builder_sequence():
    b = Sequence([Just(1), Just(2), Just(3)])
    assert b.apply({}) == [1, 2, 3]

    b = Sequence((Just(1), Just(2), Just(3)))
    assert b.apply({}) == (1, 2, 3)


def test_builder_mapping():
    b = Mapping({"a": Just(1), "b": Just(2)})
    assert b.apply({}) == {"a": 1, "b": 2}

    b = Mapping({"a": Just(1), "b": Just(2)})
    assert b.apply({}) == {"a": 1, "b": 2}


## Deferred tests


def resolve(deferred, _):
    return builder(deferred).apply({"_": _})


def test_deferred_object_are_not_hashable():
    # since __eq__ is overloaded, Deferred objects are not hashable
    with pytest.raises(TypeError, match="unhashable type"):
        hash(_.a)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ((), ()),
        ([], []),
        ({}, {}),
        ((1, 2, 3), (1, 2, 3)),
        ([1, 2, 3], [1, 2, 3]),
        ({"a": 1, "b": 2}, {"a": 1, "b": 2}),
    ],
)
def test_deferred_builds(value, expected):
    assert builder(value).apply({}) == expected


def test_deferred_supports_string_arguments():
    # deferred() is applied on all arguments of Call() except the first one and
    # sequences are transparently handled, the check far sequences was incorrect
    # for strings causing infinite recursion
    b = builder("3.14")
    assert b.apply({}) == "3.14"


def test_deferred_variable_getattr():
    v = Deferred(Variable("v"))
    p = v.copy
    assert builder(p) == Attr(v, "copy")
    assert builder(p).apply({"v": [1, 2, 3]})() == [1, 2, 3]

    p = v.copy()
    assert builder(p) == Call(Attr(v, "copy"))
    assert builder(p).apply({"v": [1, 2, 3]}) == [1, 2, 3]


class TableMock(dict):
    def __getattr__(self, attr):
        return self[attr]

    def __eq__(self, other):
        return isinstance(other, TableMock) and super().__eq__(other)


def _binop(name, switch=False):
    def method(self, other):
        if switch:
            return BinaryMock(name=name, left=other, right=self)
        else:
            return BinaryMock(name=name, left=self, right=other)

    return method


class ValueMock:
    def log(self, base=None):
        return UnaryMock(name="log", arg=base)

    def sum(self):
        return UnaryMock(name="sum", arg=self)

    def __neg__(self):
        return UnaryMock(name="neg", arg=self)

    def __invert__(self):
        return UnaryMock(name="invert", arg=self)

    __lt__ = _binop("lt")
    __gt__ = _binop("gt")
    __le__ = _binop("le")
    __ge__ = _binop("ge")
    __add__ = _binop("add")
    __radd__ = _binop("add", switch=True)
    __sub__ = _binop("sub")
    __rsub__ = _binop("sub", switch=True)
    __mul__ = _binop("mul")
    __rmul__ = _binop("mul", switch=True)
    __mod__ = _binop("mod")
    __rmod__ = _binop("mod", switch=True)
    __truediv__ = _binop("div")
    __rtruediv__ = _binop("div", switch=True)
    __floordiv__ = _binop("floordiv")
    __rfloordiv__ = _binop("floordiv", switch=True)
    __rshift__ = _binop("shift")
    __rrshift__ = _binop("shift", switch=True)
    __lshift__ = _binop("shift")
    __rlshift__ = _binop("shift", switch=True)
    __pow__ = _binop("pow")
    __rpow__ = _binop("pow", switch=True)
    __xor__ = _binop("xor")
    __rxor__ = _binop("xor", switch=True)
    __and__ = _binop("and")
    __rand__ = _binop("and", switch=True)
    __or__ = _binop("or")
    __ror__ = _binop("or", switch=True)


class ColumnMock(ValueMock):
    def __init__(self, name, dtype):
        self.name = name
        self.dtype = dtype

    def __eq__(self, other):
        return (
            isinstance(other, ColumnMock)
            and self.name == other.name
            and self.dtype == other.dtype
        )

    def __deferred_repr__(self):
        return f"<column[{self.dtype}]>"


class UnaryMock(ValueMock):
    def __init__(self, name, arg):
        self.name = name
        self.arg = arg

    def __eq__(self, other):
        return (
            isinstance(other, UnaryMock)
            and self.name == other.name
            and self.arg == other.arg
        )


class BinaryMock(ValueMock):
    def __init__(self, name, left, right):
        self.name = name
        self.left = left
        self.right = right

    def __eq__(self, other):
        return (
            isinstance(other, BinaryMock)
            and self.name == other.name
            and self.left == other.left
            and self.right == other.right
        )


@pytest.fixture
def table():
    return TableMock(
        a=ColumnMock(name="a", dtype="int"),
        b=ColumnMock(name="b", dtype="int"),
        c=ColumnMock(name="c", dtype="string"),
    )


def test_deferred_getitem(table):
    expr = _["a"]
    assert resolve(expr, table) == table["a"]
    assert repr(expr) == "$_['a']"


def test_deferred_getattr(table):
    expr = _.a
    assert resolve(expr, table) == table.a
    assert repr(expr) == "$_.a"


def test_deferred_call(table):
    expr = Deferred(Call(operator.add, _.a, 2))
    res = resolve(expr, table)
    assert res == table.a + 2
    assert repr(expr) == "add($_.a, 2)"

    func = lambda a, b: a + b
    expr = Deferred(Call(func, a=_.a, b=2))
    res = resolve(expr, table)
    assert res == table.a + 2
    assert func.__name__ in repr(expr)
    assert "a=$_.a, b=2" in repr(expr)

    expr = Deferred(Call(operator.add, (_.a, 2)))
    assert repr(expr) == "add(($_.a, 2))"


def test_deferred_method(table):
    expr = _.a.log()
    res = resolve(expr, table)
    assert res == table.a.log()
    assert repr(expr) == "$_.a.log()"


def test_deferred_method_with_args(table):
    expr = _.a.log(1)
    res = resolve(expr, table)
    assert res == table.a.log(1)
    assert repr(expr) == "$_.a.log(1)"

    expr = _.a.log(_.b)
    res = resolve(expr, table)
    assert res == table.a.log(table.b)
    assert repr(expr) == "$_.a.log($_.b)"


def test_deferred_method_with_kwargs(table):
    expr = _.a.log(base=1)
    res = resolve(expr, table)
    assert res == table.a.log(base=1)
    assert repr(expr) == "$_.a.log(base=1)"

    expr = _.a.log(base=_.b)
    res = resolve(expr, table)
    assert res == table.a.log(base=table.b)
    assert repr(expr) == "$_.a.log(base=$_.b)"


def test_deferred_apply(table):
    expr = Deferred(Call(operator.add, _.a, 2))
    res = resolve(expr, table)
    assert res == table.a + 2
    assert repr(expr) == "add($_.a, 2)"

    func = lambda a, b: a + b
    expr = Deferred(Call(func, _.a, 2))
    res = resolve(expr, table)
    assert res == table.a + 2
    assert func.__name__ in repr(expr)


@pytest.mark.parametrize(
    "symbol, op",
    [
        ("+", operator.add),
        ("-", operator.sub),
        ("*", operator.mul),
        ("/", operator.truediv),
        ("//", operator.floordiv),
        ("**", operator.pow),
        ("%", operator.mod),
        ("&", operator.and_),
        ("|", operator.or_),
        ("^", operator.xor),
        (">>", operator.rshift),
        ("<<", operator.lshift),
    ],
)
def test_deferred_binary_operations(symbol, op, table):
    expr = op(_.a, _.b)
    sol = op(table.a, table.b)
    res = resolve(expr, table)
    assert res == sol
    assert repr(expr) == f"($_.a {symbol} $_.b)"

    expr = op(1, _.a)
    sol = op(1, table.a)
    res = resolve(expr, table)
    assert res == sol
    assert repr(expr) == f"(1 {symbol} $_.a)"


@pytest.mark.parametrize(
    "sym, rsym, op",
    [
        ("==", "==", operator.eq),
        ("!=", "!=", operator.ne),
        ("<", ">", operator.lt),
        ("<=", ">=", operator.le),
        (">", "<", operator.gt),
        (">=", "<=", operator.ge),
    ],
)
def test_deferred_compare_operations(sym, rsym, op, table):
    expr = op(_.a, _.b)
    sol = op(table.a, table.b)
    res = resolve(expr, table)
    assert res == sol
    assert repr(expr) == f"($_.a {sym} $_.b)"

    expr = op(1, _.a)
    sol = op(1, table.a)
    res = resolve(expr, table)
    assert res == sol
    assert repr(expr) == f"($_.a {rsym} 1)"


@pytest.mark.parametrize(
    "symbol, op",
    [
        ("-", operator.neg),
        ("~", operator.invert),
    ],
)
def test_deferred_unary_operations(symbol, op, table):
    expr = op(_.a)
    sol = op(table.a)
    res = resolve(expr, table)
    assert res == sol
    assert repr(expr) == f"{symbol}$_.a"


@pytest.mark.parametrize("obj", [_, _.a, _.a.b[0]])
def test_deferred_is_not_iterable(obj):
    with pytest.raises(TypeError, match="object is not iterable"):
        sorted(obj)

    with pytest.raises(TypeError, match="object is not iterable"):
        iter(obj)

    with pytest.raises(TypeError, match="is not an iterator"):
        next(obj)


@pytest.mark.parametrize("obj", [_, _.a, _.a.b[0]])
def test_deferred_is_not_truthy(obj):
    with pytest.raises(
        TypeError, match="The truth value of Deferred objects is not defined"
    ):
        bool(obj)
